"""
Halo Mass Function and supporting functions.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import math

from yt.funcs import mylog
from yt.units.yt_array import \
    YTArray, \
    YTQuantity
from yt.utilities.physical_ratios import \
    rho_crit_g_cm3_h2

class HaloMassFcn():
    r"""
    Initialize a HaloMassFcn object to analyze the distribution of halos as 
    a function of mass.  A mass function can be created for a set of 
    simulated halos, an analytic fit to can be created for a redshift and 
    set of cosmological parameters, or both can be created.

    Provided with a halo dataset object, this will make a the mass function 
    for simulated halos.  Providing a simulation dataset will set as many 
    of the cosmological parameters as possible for the creation of the 
    analytic mass function.

    The HaloMassFcn object has arrays hanging off of it containing the mass
    function information.

    masses_sim : Array 
        Halo masses from simulated halos. Units: M_solar.
    n_cumulative_sim : Array
        Number density of halos with mass greater than the corresponding 
        mass in masses_sim (simulated). Units: comoving Mpc^-3
    masses_analytic : Array
        Masses used for the generation of the analytic mass function, Units:
        M_solar.
    n_cumulative_analytic : Array
        Number density of halos with mass greater then the corresponding
        mass in masses_analytic (analytic). Units: comoving Mpc^-3
    dndM_dM_analytic : Array
        Differential number density of halos, (dn/dM)*dM (analytic).

    The HaloMassFcn object also has a convenience function write_out() that
    will write out the data to disk.

    Creating a HaloMassFcn object with no arguments will produce an analytic
    mass function at redshift = 0 using default cosmological values.

    Parameters
    ----------
    simulation_ds : Simulation dataset object
        The loaded simulation dataset, used to set cosmological parameters.
        Default : None.
    halos_ds : Halo dataset object
        The halos from a simulation to be used for creation of the 
        halo mass function in the simulation.
        Default : None.
    make_analytic : bool 
        Whether or not to calculate the analytic mass function to go with 
        the simulated halo mass function.  Automatically set to true if a 
        simulation dataset is provided.
        Default : True.
    omega_matter0 : float
        The fraction of the universe made up of matter (dark and baryonic). 
        Default : 0.2726.
    omega_lambda0 : float
        The fraction of the universe made up of dark energy. 
        Default : 0.7274.
    omega_baryon0  : float 
        The fraction of the universe made up of baryonic matter. This is not 
        always stored in the dataset and should be checked by hand.
        Default : 0.0456.
    hubble0 : float 
        The expansion rate of the universe in units of 100 km/s/Mpc. 
        Default : 0.704.
    sigma8 : float 
        The amplitude of the linear power spectrum at z=0 as specified by 
        the rms amplitude of mass-fluctuations in a top-hat sphere of radius 
        8 Mpc/h. This is not always stored in the dataset and should be 
        checked by hand.
        Default : 0.86.
    primoridal_index : float 
        This is the index of the mass power spectrum before modification by 
        the transfer function. A value of 1 corresponds to the scale-free 
        primordial spectrum. This is not always stored in the dataset and 
        should be checked by hand.
        Default : 1.0.
    this_redshift : float 
        The current redshift. 
        Default : 0.
    log_mass_min : float 
        The log10 of the mass of the minimum of the halo mass range. This is
        set automatically by the range of halo masses if a simulated halo 
        dataset is provided. If a halo dataset if not provided and no value
        is specified, it will be set to 5. Units: M_solar
        Default : None.
    log_mass_max : float 
        The log10 of the mass of the maximum of the halo mass range. This is
        set automatically by the range of halo masses if a simulated halo 
        dataset is provided. If a halo dataset if not provided and no value
        is specified, it will be set to 16. Units: M_solar
        Default : None.
    num_sigma_bins : float
        The number of bins (points) to use for the calculation of the 
        analytic mass function. 
        Default : 360.
    fitting_function : int
        Which fitting function to use. 1 = Press-Schechter, 2 = Jenkins, 
        3 = Sheth-Tormen, 4 = Warren, 5 = Tinker
        Default : 4.

    Examples
    --------

    This creates the halo mass function for a halo dataset from a simulation
    and the analytic mass function at the same redshift as the dataset,
    using as many cosmological parameters as can be pulled from the dataset.

    >>> halos_ds = load("rockstar_halos/halo_0.0.bin")
    >>> hmf = HaloMassFcn(halos_ds=halos_ds)
    >>> plt.loglog(hmf.masses_sim, hmf.n_cumulative_sim)
    >>> plt.loglog(hmf.masses_analytic, hmf.n_cumulative_analytic)
    >>> plt.savefig("mass_function.png")

    This creates only the analytic halo mass function for a simulation
    dataset, with default values for cosmological parameters not stored in 
    the dataset.

    >>> ds = load("enzo_tiny_cosmology/DD0046/DD0046")
    >>> hmf = HaloMassFcn(simulation_ds=ds)
    >>> plt.loglog(hmf.masses_analytic, hmf.n_cumulative_analytic)
    >>> plt.savefig("mass_function.png")
    
    This creates the analytic mass function for an arbitrary set of 
    cosmological parameters, with neither a simulation nor halo dataset.

    >>> hmf = HaloMassFcn(omega_baryon0=0.05, omega_matter0=0.27, 
                          omega_lambda0=0.73, hubble0=0.7, this_redshift=10,
                          log_mass_min=5, log_mass_max=9)
    >>> plt.loglog(hmf.masses_analytic, hmf.n_cumulative_analytic)
    >>> plt.savefig("mass_function.png")
    """
    def __init__(self, simulation_ds=None, halos_ds=None, make_analytic=True, 
    omega_matter0=0.2726, omega_lambda0=0.7274, omega_baryon0=0.0456, hubble0=0.704, 
    sigma8=0.86, primordial_index=1.0, this_redshift=0, log_mass_min=None, 
    log_mass_max=None, num_sigma_bins=360, fitting_function=4):
        self.simulation_ds = simulation_ds
        self.halos_ds = halos_ds
        self.omega_matter0 = omega_matter0
        self.omega_lambda0 = omega_lambda0
        self.omega_baryon0 = omega_baryon0
        self.hubble0 = hubble0
        self.sigma8 = sigma8
        self.primordial_index = primordial_index
        self.this_redshift = this_redshift
        self.log_mass_min = log_mass_min
        self.log_mass_max = log_mass_max
        self.num_sigma_bins = num_sigma_bins
        self.fitting_function = fitting_function
        self.make_analytic = make_analytic
        self.make_simulated = False
        """
        If we want to make an analytic mass function, grab what we can from either the 
        halo file or the data set, and make sure that the user supplied everything else
        that is needed.
        """
        # If we don't have any datasets, make the analytic function with user values
        if simulation_ds is None and halos_ds is None:
            # Set a reasonable mass min and max if none were provided
            if log_mass_min is None:
                self.log_mass_min = 5
            if log_mass_max is None:
                self.log_mass_max = 16
        # If we're making the analytic function...
        if self.make_analytic is True:
            # Try to set cosmological parameters from the simulation dataset
            if simulation_ds is not None:
                self.omega_matter0 = self.simulation_ds.omega_matter
                self.omega_lambda0 = self.simulation_ds.omega_lambda
                self.hubble0 = self.simulation_ds.hubble_constant
                self.this_redshift = self.simulation_ds.current_redshift
                # Set a reasonable mass min and max if none were provided
                if log_mass_min is None:
                    self.log_mass_min = 5
                if log_mass_max is None:
                    self.log_mass_max = 16
            # If we have a halo dataset but not a simulation dataset, use that instead
            if simulation_ds is None and halos_ds is not None:
                self.omega_matter0 = self.halos_ds.omega_matter
                self.omega_lambda0 = self.halos_ds.omega_lambda
                self.hubble0 = self.halos_ds.hubble_constant
                self.this_redshift = self.halos_ds.current_redshift
                # If the user didn't specify mass min and max, set them from the halos
                if log_mass_min is None:
                    self.set_mass_from_halos("min_mass")
                if log_mass_max is None:
                    self.set_mass_from_halos("max_mass")
            # Do the calculations.
            self.sigmaM()
            self.dndm()
            # Return the mass array in M_solar rather than M_solar/h
            self.masses_analytic = YTArray(self.masses_analytic/self.hubble0, "Msun")
            # The halo arrays will already have yt units, but the analytic forms do 
            # not. If a dataset has been provided, use that to give them units. At the
            # same time, convert to comoving (Mpc)^-3
            if simulation_ds is not None:
                self.n_cumulative_analytic = simulation_ds.arr(self.n_cumulative_analytic, 
                                                          "(Mpccm)**(-3)")
            elif halos_ds is not None:
                self.n_cumulative_analytic = halos_ds.arr(self.n_cumulative_analytic, 
                                                          "(Mpccm)**(-3)")
            else:
                from yt.units.unit_registry import UnitRegistry
                from yt.units.dimensions import length
                hmf_registry = UnitRegistry()
                for my_unit in ["m", "pc", "AU", "au"]:
                    new_unit = "%scm" % my_unit
                    hmf_registry.add(new_unit, 
                                     hmf_registry.lut[my_unit][0] / 
                                     (1 + self.this_redshift),
                                     length, "\\rm{%s}/(1+z)" % my_unit)                         
                self.n_cumulative_analytic = YTArray(self.n_cumulative_analytic, 
                                                     "(Mpccm)**(-3)", 
                                                     registry=hmf_registry) 


        """
        If a halo file has been supplied, make a mass function for the simulated halos.
        """
        if halos_ds is not None:
            # Used to check if a simulated halo mass function exists to write out
            self.make_simulated=True
            # Calculate the simulated halo mass function
            self.create_sim_hmf()

    """
    If we're making an analytic fit and have a halo dataset, but don't have log_mass_min 
    or log_mass_max from the user, set it from the range of halo masses.
    """
    def set_mass_from_halos(self, which_limit):
        data_source = self.halos_ds.all_data()
        if which_limit == "min_mass":
            self.log_mass_min = \
              int(np.log10(np.amin(data_source["particle_mass"].in_units("Msun"))))
        if which_limit == "max_mass":
            self.log_mass_max = \
              int(np.log10(np.amax(data_source["particle_mass"].in_units("Msun"))))+1
    
    """
    Here's where we create the halo mass functions from simulated halos
    """
    def create_sim_hmf(self):
        data_source = self.halos_ds.all_data()
        # We're going to use indices to count the number of halos above a given mass
        masses_sim = np.sort(data_source["particle_mass"].in_units("Msun"))
        # Determine the size of the simulation volume in comoving Mpc**3
        sim_volume = self.halos_ds.domain_width.in_units('Mpccm').prod()
        n_cumulative_sim = np.arange(len(masses_sim),0,-1)
        # We don't want repeated halo masses, and the unique indices tell us which values 
        # correspond to distinct halo masses.
        self.masses_sim, unique_indices = np.unique(masses_sim, return_index=True)
        # Now make this an actual number density of halos as a function of mass.
        self.n_cumulative_sim = n_cumulative_sim[unique_indices]/sim_volume
        # masses_sim and n_cumulative_sim are now set, but remember that the log10 quantities
        # are what is usually plotted for a halo mass function.

    def write_out(self, prefix='HMF', analytic=True, simulated=True):
        """
        Writes out the halo mass functions to file(s) with prefix *prefix*.
        """
        # First the analytic file, check that analytic fit exists and was requested
        if analytic:
            if self.make_analytic:
                fitname = prefix + '-analytic.dat'
                fp = open(fitname, "w")
                line = \
                "#Columns:\n" + \
                "#1. mass (M_solar)\n" + \
                "#2. cumulative number density of halos [comoving Mpc^-3]\n" + \
                "#3. (dn/dM)*dM (differential number density of halos) [comoving Mpc^-3]\n"
                fp.write(line)
                for i in range(self.masses_analytic.size - 1):
                    line = "%e\t%e\t%e\n" % (self.masses_analytic[i],
                    self.n_cumulative_analytic[i], 
                    self.dndM_dM_analytic[i])
                    fp.write(line)
                fp.close()
            # If the analytic halo mass function wasn't created, warn the user
            else:
                mylog.warning("The analytic halo mass function was not created and cannot be " +
                              "written out! Specify its creation with " +
                              "HaloMassFcn(make_analytic=True, other_args) when creating the " +
                              "HaloMassFcn object.")
        # Write out the simulated mass fucntion if it exists and was requested
        if simulated:
            if self.make_simulated:
                haloname = prefix + '-simulated.dat'
                fp = open(haloname, "w")
                line = \
                "#Columns:\n" + \
                "#1. mass [Msun]\n" + \
                "#2. cumulative number density of halos [comoving Mpc^-3]\n"
                fp.write(line)
                for i in range(self.masses_sim.size - 1):
                    line = "%e\t%e\n" % (self.masses_sim[i], 
                    self.n_cumulative_sim[i])
                    fp.write(line)
                fp.close()
            # If the simulated halo mass function wasn't created, warn the user
            else:
                mylog.warning("The simulated halo mass function was not created and cannot " +
                              "be written out! Specify its creation by providing a loaded " +
                              "halo dataset with HaloMassFcn(ds_halos=loaded_halo_dataset, " +
                              "other_args) when creating the HaloMassFcn object.")

    def sigmaM(self):
        """
        Written by BWO, 2006 (updated 25 January 2007).
        Converted to Python by Stephen Skory December 2009.

        This routine takes in cosmological parameters and creates a file (array) with
        sigma(M) in it, which is necessary for various press-schechter type
        stuff.  In principle one can calculate it ahead of time, but it's far,
        far faster in the long run to calculate your sigma(M) ahead of time.
        
        Inputs: cosmology, user must set parameters
        
        Outputs: four columns of data containing the following information:

        1) mass (Msolar/h)
        2) sigma (normalized) using Msun/h as the input
        
        The arrays output are used later.
        """
        
        # Set up the transfer function object.
        self.TF = TransferFunction(self.omega_matter0, self.omega_baryon0, 0.0, 0,
            self.omega_lambda0, self.hubble0, self.this_redshift);

        if self.TF.qwarn:
            mylog.error("You should probably fix your cosmology parameters!")

        # output arrays
        # 1) mass (M_solar/h), changed to M_solar/h at output
        self.masses_analytic = np.empty(self.num_sigma_bins, dtype='float64')
        # 2) sigma(M, z=0, where mass is in Msun/h)
        self.sigmaarray = np.empty(self.num_sigma_bins, dtype='float64')

        # get sigma_8 normalization
        R = 8.0;  # in units of Mpc/h (comoving)

        sigma8_unnorm = math.sqrt(self.sigma_squared_of_R(R));
        sigma_normalization = self.sigma8 / sigma8_unnorm;

        # rho0 in units of h^2 Msolar/Mpc^3
        rho0 = YTQuantity(self.omega_matter0 * rho_crit_g_cm3_h2 * self.hubble0**2,
                          'g/cm**3').in_units('Msun/Mpc**3')
        rho0 = rho0.value.item()       

        # spacing in mass of our sigma calculation
        dm = (float(self.log_mass_max) - self.log_mass_min)/self.num_sigma_bins;

        """
         loop over the total number of sigma_bins the user has requested. 
         For each bin, calculate mass and equivalent radius, and call
         sigma_squared_of_R to get the sigma(R) (equivalent to sigma(M)),
         normalize by user-specified sigma_8, and then write out.
        """
        for i in range(self.num_sigma_bins):
    
            # thislogmass is in units of Msolar, NOT Msolar/h
            thislogmass = self.log_mass_min +  i*dm
    
            # mass in units of h^-1 Msolar
            thismass = math.pow(10.0, thislogmass) * self.hubble0; 
    
            # radius is in units of h^-1 Mpc (comoving)
            thisradius = math.pow( 3.0*thismass / 4.0 / math.pi / rho0, 1.0/3.0 );
    
            R = thisradius; # h^-1 Mpc (comoving)
    
            self.masses_analytic[i] = thismass;  # Msun/h
    
            # get normalized sigma(R)
            self.sigmaarray[i] = math.sqrt(self.sigma_squared_of_R(R)) * sigma_normalization;
            # All done!

    def dndm(self):
        # constants - set these before calling any functions!
        # rho0 in units of h^2 Msolar/Mpc^3
        rho0 = YTQuantity(self.omega_matter0 * rho_crit_g_cm3_h2 * self.hubble0**2, 
                          'g/cm**3').in_units('Msun/Mpc**3')
        rho0 = rho0.value.item()

        self.delta_c0 = 1.69;  # critical density for turnaround (Press-Schechter)
        
        n_cumulative_analytic = 0.0;  # keep track of cumulative number density
        
        # Loop over masses, going BACKWARD, and calculate dn/dm as well as the 
        # cumulative mass function.
        
        # output arrays
        # 5) (dn/dM)*dM (differential number density of halos, per Mpc^3 (NOT h^3/Mpc^3)
        self.dndM_dM_analytic = np.empty(self.num_sigma_bins, dtype='float64')
        # 6) cumulative number density of halos (per Mpc^3, NOT h^3/Mpc^3)
        self.n_cumulative_analytic = np.zeros(self.num_sigma_bins, dtype='float64')
        
        for j in range(self.num_sigma_bins - 1):
            i = (self.num_sigma_bins - 2) - j
        
            thissigma = self.sigmaof_M_z(i, self.this_redshift);
            nextsigma = self.sigmaof_M_z(i+1, self.this_redshift);
            
            # calc dsigmadm - has units of h (since masses_analytic has units of h^-1)
            dsigmadm = (nextsigma-thissigma) / (self.masses_analytic[i+1] - self.masses_analytic[i]);

            # calculate dn(M,z) (dn/dM * dM)
            # this has units of h^3 since rho0 has units of h^2, dsigmadm
            # has units of h, and masses_analytic has units of h^-1
            dndM_dM_analytic = -1.0 / thissigma * dsigmadm * rho0 / self.masses_analytic[i] * \
            self.multiplicityfunction(thissigma)*(self.masses_analytic[i+1] - self.masses_analytic[i]);

            # scale by h^3 to get rid of all factors of h
            dndM_dM_analytic *= math.pow(self.hubble0, 3.0);
            
            # keep track of cumulative number density
            if dndM_dM_analytic > 1.0e-20:
                n_cumulative_analytic += dndM_dM_analytic;
            
            # Store this.
            self.n_cumulative_analytic[i] = n_cumulative_analytic
            self.dndM_dM_analytic[i] = dndM_dM_analytic
        

    def sigma_squared_of_R(self, R):
        """
        calculates sigma^2(R).  This is the routine where the magic happens (or
        whatever it is that we do here).  Integrates the sigma_squared_integrand
        parameter from R to infinity.  Calls GSL (gnu scientific library) to do
        the actual integration.  
        
        Note that R is in h^-1 Mpc (comoving)
        """
        self.R = R
        result = integrate_inf(self.sigma_squared_integrand)

        sigmasquaredofR = result / 2.0 / math.pi / math.pi

        return sigmasquaredofR;

    def sigma_squared_integrand(self, k):
        """
        integrand for integral to get sigma^2(R).
        """

        Rcom = self.R;  # this is R in comoving Mpc/h

        f = k*k*self.PofK(k)*np.power( abs(self.WofK(Rcom,k)), 2.0);

        return f

    def PofK(self, k):
        """
        returns power spectrum as a function of wavenumber k
        """

        thisPofK = np.power(k, self.primordial_index) * np.power( self.TofK(k), 2.0);

        return thisPofK;

    def TofK(self, k):
        """
        returns transfer function as a function of wavenumber k.
        """
        
        thisTofK = self.TF.TFmdm_onek_hmpc(k);

        return thisTofK;

    def WofK(self, R, k):
        """
        returns W(k), which is the fourier transform of the top-hat function.
        """

        x = R*k;

        thisWofK = 3.0 * ( np.sin(x) - x*np.cos(x) ) / (x*x*x);

        return thisWofK;

    def multiplicityfunction(self, sigma):
        """
        Multiplicity function - this is where the various fitting functions/analytic 
        theories are different.  The various places where I found these fitting functions
        are listed below.
        """
        
        nu = self.delta_c0 / sigma;
        
        if self.fitting_function==1:
            # Press-Schechter (This form from Jenkins et al. 2001, MNRAS 321, 372-384, eqtn. 5)
            thismult = math.sqrt(2.0/math.pi) * nu * math.exp(-0.5*nu*nu);
        
        elif self.fitting_function==2:
            # Jenkins et al. 2001, MNRAS 321, 372-384, eqtn. 9
            thismult = 0.315 * math.exp( -1.0 * math.pow( abs( math.log(1.0/sigma) + 0.61), 3.8 ) );
        
        elif self.fitting_function==3:
            # Sheth-Tormen 1999, eqtn 10, using expression from Jenkins et al. 2001, eqtn. 7
            A=0.3222;
            a=0.707;
            p=0.3;
            thismult = A*math.sqrt(2.0*a/math.pi)*(1.0+ math.pow( 1.0/(nu*nu*a), p) )*\
            nu * math.exp(-0.5*a*nu*nu);
        
        elif self.fitting_function==4:
            # LANL fitting function - Warren et al. 2005, astro-ph/0506395, eqtn. 5 
            A=0.7234; 
            a=1.625; 
            b=0.2538; 
            c=1.1982;
            thismult = A*( math.pow(sigma, -1.0*a) + b)*math.exp(-1.0*c / sigma / sigma );

        elif self.fitting_function==5:
            # Tinker et al. 2008, eqn 3, \Delta=300 # \Delta=200
            A = 0.2 #0.186
            a = 1.52 #1.47
            b = 2.25 #2.57
            c = 1.27 #1.19
            thismult = A * ( math.pow((sigma / b), -a) + 1) * \
                math.exp(-1 * c / sigma / sigma)
        
        else:
            mylog.error("Don't understand this.  Fitting function requested is %d\n",
            self.fitting_function)
            return None
        
        return thismult

    def sigmaof_M_z(self, sigmabin, redshift):
        """
        sigma(M, z)
        """
        
        thissigma = self.Dofz(redshift) * self.sigmaarray[sigmabin];
        
        return thissigma;

    def Dofz(self, redshift):
        """
        Growth function
        """

        thisDofz = self.gofz(redshift) / self.gofz(0.0) / (1.0+redshift);

        return thisDofz;


    def gofz(self, redshift):
        """
        g(z) - I don't think this has any other name
        """

        thisgofz = 2.5 * self.omega_matter_of_z(redshift) / \
        ( math.pow( self.omega_matter_of_z(redshift), 4.0/7.0 ) - \
          self.omega_lambda_of_z(redshift) + \
          ( (1.0 + self.omega_matter_of_z(redshift) / 2.0) * \
          (1.0 + self.omega_lambda_of_z(redshift) / 70.0) ))

        return thisgofz;


    def omega_matter_of_z(self,redshift):
        """
        Omega matter as a function of redshift
        """
        
        thisomofz = self.omega_matter0 * math.pow( 1.0+redshift, 3.0) / \
            math.pow( self.Eofz(redshift), 2.0 );
        
        return thisomofz;

    def omega_lambda_of_z(self,redshift):
        """
        Omega lambda as a function of redshift
        """

        thisolofz = self.omega_lambda0 / math.pow( self.Eofz(redshift), 2.0 )

        return thisolofz;

    def Eofz(self, redshift):
        """
        E(z) - I don't think this has any other name
        """
        thiseofz = math.sqrt( self.omega_lambda0 \
            + (1.0 - self.omega_lambda0 - self.omega_matter0)*math.pow( 1.0+redshift, 2.0) \
            + self.omega_matter0 * math.pow( 1.0+redshift, 3.0)  );

        return thiseofz;


""" 
Fitting Formulae for CDM + Baryon + Massive Neutrino (MDM) cosmologies.
Daniel J. Eisenstein & Wayne Hu, Institute for Advanced Study

There are two primary routines here, one to set the cosmology, the
other to construct the transfer function for a single wavenumber k. 
You should call the former once (per cosmology) and the latter as 
many times as you want. 

   TFmdm_set_cosm() -- User passes all the cosmological parameters as
   arguments; the routine sets up all of the scalar quantities needed 
   computation of the fitting formula.  The input parameters are: 
   1) omega_matter -- Density of CDM, baryons, and massive neutrinos,
                      in units of the critical density. 
   2) omega_baryon -- Density of baryons, in units of critical. 
   3) omega_hdm    -- Density of massive neutrinos, in units of critical 
   4) degen_hdm    -- (Int) Number of degenerate massive neutrino species 
   5) omega_lambda -- Cosmological constant 
   6) hubble       -- Hubble constant, in units of 100 km/s/Mpc 
   7) redshift     -- The redshift at which to evaluate */

   TFmdm_onek_mpc() -- User passes a single wavenumber, in units of Mpc^-1.
   Routine returns the transfer function from the Eisenstein & Hu
   fitting formula, based on the cosmology currently held in the 
   internal variables.  The routine returns T_cb (the CDM+Baryon
   density-weighted transfer function), although T_cbn (the CDM+
   Baryon+Neutrino density-weighted transfer function) is stored
   in the global variable tf_cbnu. */

We also supply TFmdm_onek_hmpc(), which is identical to the previous
routine, but takes the wavenumber in units of h Mpc^-1.

We hold the internal scalar quantities in global variables, so that
the user may access them in an external program, via "extern" declarations.

Please note that all internal length scales are in Mpc, not h^-1 Mpc!
"""

class TransferFunction(object):
    """
    This routine takes cosmological parameters and a redshift and sets up
    all the internal scalar quantities needed to compute the transfer function.

    Parameters
    ----------
    omega_matter : float
        Density of CDM, baryons, and massive neutrinos, in units 
        of the critical density.
    omega_baryon : float
        Density of baryons, in units of critical.
    omega_hdm : float
        Density of massive neutrinos, in units of critical
    degen_hdm : integer
        Number of degenerate massive neutrino species
    omega_lambda : float
        Cosmological constant
    hubble : float
        Hubble constant, in units of 100 km/s/Mpc
    redshift : float
        The redshift at which to evaluate

    Returns 0 if all is well, 1 if a warning was issued. Otherwise,
    sets many global variables for use in TFmdm_onek_mpc()
    """
    def __init__(self, omega_matter, omega_baryon, omega_hdm,
                 degen_hdm, omega_lambda, hubble, redshift):
        self.qwarn = 0;
        self.theta_cmb = 2.728/2.7 # Assuming T_cmb = 2.728 K
    
        # Look for strange input
        if (omega_baryon<0.0):
            mylog.error("TFmdm_set_cosm(): Negative omega_baryon set to trace amount.\n")
            self.qwarn = 1
        if (omega_hdm<0.0):
            mylog.error("TFmdm_set_cosm(): Negative omega_hdm set to trace amount.\n")
            self.qwarn = 1;
        if (hubble<=0.0):
            mylog.error("TFmdm_set_cosm(): Negative Hubble constant illegal.\n")
            return None
        elif (hubble>2.0):
            mylog.error("TFmdm_set_cosm(): Hubble constant should be in units of 100 km/s/Mpc.\n");
            self.qwarn = 1;
        if (redshift<=-1.0):
            mylog.error("TFmdm_set_cosm(): Redshift < -1 is illegal.\n");
            return None
        elif (redshift>99.0):
            mylog.error("TFmdm_set_cosm(): Large redshift entered.  TF may be inaccurate.\n");
            self.qwarn = 1;

        if (degen_hdm<1): degen_hdm=1;
        self.num_degen_hdm = degen_hdm;	
        # Have to save this for TFmdm_onek_mpc()
        # This routine would crash if baryons or neutrinos were zero,
        # so don't allow that.
        if (omega_baryon<=0): omega_baryon=1e-5;
        if (omega_hdm<=0): omega_hdm=1e-5;
    
        self.omega_curv = 1.0-omega_matter-omega_lambda;
        self.omhh = omega_matter*SQR(hubble);
        self.obhh = omega_baryon*SQR(hubble);
        self.onhh = omega_hdm*SQR(hubble);
        self.f_baryon = omega_baryon/omega_matter;
        self.f_hdm = omega_hdm/omega_matter;
        self.f_cdm = 1.0-self.f_baryon-self.f_hdm;
        self.f_cb = self.f_cdm+self.f_baryon;
        self.f_bnu = self.f_baryon+self.f_hdm;
    
        # Compute the equality scale.
        self.z_equality = 25000.0*self.omhh/SQR(SQR(self.theta_cmb)) # Actually 1+z_eq
        self.k_equality = 0.0746*self.omhh/SQR(self.theta_cmb);
    
        # Compute the drag epoch and sound horizon
        z_drag_b1 = 0.313*math.pow(self.omhh,-0.419)*(1+0.607*math.pow(self.omhh,0.674));
        z_drag_b2 = 0.238*math.pow(self.omhh,0.223);
        self.z_drag = 1291*math.pow(self.omhh,0.251)/(1.0+0.659*math.pow(self.omhh,0.828))* \
            (1.0+z_drag_b1*math.pow(self.obhh,z_drag_b2));
        self.y_drag = self.z_equality/(1.0+self.z_drag);
    
        self.sound_horizon_fit = 44.5*math.log(9.83/self.omhh)/math.sqrt(1.0+10.0*math.pow(self.obhh,0.75));
    
        # Set up for the free-streaming & infall growth function 
        self.p_c = 0.25*(5.0-math.sqrt(1+24.0*self.f_cdm));
        self.p_cb = 0.25*(5.0-math.sqrt(1+24.0*self.f_cb));
    
        omega_denom = omega_lambda+SQR(1.0+redshift)*(self.omega_curv+\
                omega_matter*(1.0+redshift));
        self.omega_lambda_z = omega_lambda/omega_denom;
        self.omega_matter_z = omega_matter*SQR(1.0+redshift)*(1.0+redshift)/omega_denom;
        self.growth_k0 = self.z_equality/(1.0+redshift)*2.5*self.omega_matter_z/ \
            (math.pow(self.omega_matter_z,4.0/7.0)-self.omega_lambda_z+ \
            (1.0+self.omega_matter_z/2.0)*(1.0+self.omega_lambda_z/70.0));
        self.growth_to_z0 = self.z_equality*2.5*omega_matter/(math.pow(omega_matter,4.0/7.0) \
            -omega_lambda + (1.0+omega_matter/2.0)*(1.0+omega_lambda/70.0));
        self.growth_to_z0 = self.growth_k0/self.growth_to_z0;	
        
        # Compute small-scale suppression
        self.alpha_nu = self.f_cdm/self.f_cb*(5.0-2.*(self.p_c+self.p_cb))/(5.-4.*self.p_cb)* \
        math.pow(1+self.y_drag,self.p_cb-self.p_c)* \
        (1+self.f_bnu*(-0.553+0.126*self.f_bnu*self.f_bnu))/ \
        (1-0.193*math.sqrt(self.f_hdm*self.num_degen_hdm)+0.169*self.f_hdm*math.pow(self.num_degen_hdm,0.2))* \
        (1+(self.p_c-self.p_cb)/2*(1+1/(3.-4.*self.p_c)/(7.-4.*self.p_cb))/(1+self.y_drag));
        self.alpha_gamma = math.sqrt(self.alpha_nu);
        self.beta_c = 1/(1-0.949*self.f_bnu);
        # Done setting scalar variables
        self.hhubble = hubble # Need to pass Hubble constant to TFmdm_onek_hmpc()
        

    def TFmdm_onek_mpc(self,  kk):
        """
        Given a wavenumber in Mpc^-1, return the transfer function for the
        cosmology held in the global variables.

        Parameters
        ----------
        kk : float
            Wavenumber in Mpc^-1

        Returns
        -------
        growth_cb : float
            the transfer function for density-weighted
            CDM + Baryon perturbations. (returned and set as a global var)
        growth_cbnu : float
            the transfer function for density-weighted
            CDM + Baryon + Massive Neutrino perturbations.
            (set as a global var)
        """
    
        self.qq = kk/self.omhh*SQR(self.theta_cmb);
    
        # Compute the scale-dependent growth functions
        self.y_freestream = 17.2*self.f_hdm*(1+0.488*math.pow(self.f_hdm,-7.0/6.0))* \
            SQR(self.num_degen_hdm*self.qq/self.f_hdm);
        temp1 = math.pow(self.growth_k0, 1.0-self.p_cb);
        temp2 = np.power(self.growth_k0/(1+self.y_freestream),0.7);
        self.growth_cb = np.power(1.0+temp2, self.p_cb/0.7)*temp1;
        self.growth_cbnu = np.power(np.power(self.f_cb,0.7/self.p_cb)+temp2, self.p_cb/0.7)*temp1;
    
        # Compute the master function
        self.gamma_eff = self.omhh*(self.alpha_gamma+(1-self.alpha_gamma)/ \
            (1+SQR(SQR(kk*self.sound_horizon_fit*0.43))));
        self.qq_eff = self.qq*self.omhh/self.gamma_eff;
    
        tf_sup_L = np.log(2.71828+1.84*self.beta_c*self.alpha_gamma*self.qq_eff);
        tf_sup_C = 14.4+325/(1+60.5*np.power(self.qq_eff,1.11));
        self.tf_sup = tf_sup_L/(tf_sup_L+tf_sup_C*SQR(self.qq_eff));
    
        self.qq_nu = 3.92*self.qq*math.sqrt(self.num_degen_hdm/self.f_hdm);
        self.max_fs_correction = 1+1.2*math.pow(self.f_hdm,0.64)*math.pow(self.num_degen_hdm,0.3+0.6*self.f_hdm)/ \
            (np.power(self.qq_nu,-1.6)+np.power(self.qq_nu,0.8));
        self.tf_master = self.tf_sup*self.max_fs_correction;
    
        # Now compute the CDM+HDM+baryon transfer functions
        tf_cb = self.tf_master*self.growth_cb/self.growth_k0;
        #tf_cbnu = self.tf_master*self.growth_cbnu/self.growth_k0;
        return tf_cb


    def TFmdm_onek_hmpc(self, kk):
        """
        Given a wavenumber in h Mpc^-1, return the transfer function for the
        cosmology held in the global variables.

        Parameters
        ----------
        kk : float
            Wavenumber in h Mpc^-1

        Returns
        -------
        growth_cb : float
            the transfer function for density-weighted
            CDM + Baryon perturbations. (return and set as a global var) 
        growth_cbnu : float
            the transfer function for density-weighted
            CDM + Baryon + Massive Neutrino perturbations.
        """
        return self.TFmdm_onek_mpc(kk*self.hhubble);

def SQR(a):
    return a*a

def integrate_inf(fcn, error=1e-3, initial_guess=10):
    """
    Integrate a function *fcn* from zero to infinity, stopping when the answer
    changes by less than *error*. Hopefully someday we can do something
    better than this!
    """
    xvals = np.logspace(0,np.log10(initial_guess), initial_guess+1)-.9
    yvals = fcn(xvals)
    xdiffs = xvals[1:] - xvals[:-1]
    # Trapezoid rule, but with different dxes between values, so np.trapz
    # will not work.
    areas = (yvals[1:] + yvals[:-1]) * xdiffs / 2.0
    area0 = np.sum(areas)
    # Next guess.
    next_guess = 10 * initial_guess
    xvals = np.logspace(0,np.log10(next_guess), 2*initial_guess**2+1)-.99
    yvals = fcn(xvals)
    xdiffs = xvals[1:] - xvals[:-1]
    # Trapezoid rule.
    areas = (yvals[1:] + yvals[:-1]) * xdiffs / 2.0
    area1 = np.sum(areas)
    # Now we refine until the error is smaller than *error*.
    diff = area1 - area0
    area_last = area1
    one_pow = 3
    while diff > error:
        next_guess *= 10
        xvals = np.logspace(0,np.log10(next_guess), one_pow*initial_guess**one_pow+1) - (1 - 0.1**one_pow)
        yvals = fcn(xvals)
        xdiffs = xvals[1:] - xvals[:-1]
        # Trapezoid rule.
        areas = (yvals[1:] + yvals[:-1]) * xdiffs / 2.0
        area_next = np.sum(areas)
        diff = area_next - area_last
        area_last = area_next
        one_pow+=1
    return area_last



import numpy as np

from yt.utilities.lib.particle_mesh_operations import \
    CICSample_3
from yt.testing import assert_allclose

def setup():
    pass

def test_sample() :

    grid = {}

    dims = np.array([64,64,64], dtype='int32')
    
    inds = np.indices(dims)
    grid["x"] = inds[0] + 0.5
    grid["y"] = inds[1] + 0.5
    grid["z"] = inds[2] + 0.5

    num_particles = np.int64(1000)
    
    xp = np.random.uniform(low=1.0, high=63.0, size=num_particles)
    yp = np.random.uniform(low=1.0, high=63.0, size=num_particles)
    zp = np.random.uniform(low=1.0, high=63.0, size=num_particles)

    xfield = np.zeros((num_particles))
    yfield = np.zeros((num_particles))
    zfield = np.zeros((num_particles))

    dx = 1.
    le = np.zeros((3))
    
    CICSample_3(xp,yp,zp,xfield,num_particles,grid["x"],
                le,dims,dx)
    CICSample_3(xp,yp,zp,yfield,num_particles,grid["y"],
                le,dims,dx)
    CICSample_3(xp,yp,zp,zfield,num_particles,grid["z"],
                le,dims,dx)

    assert_allclose(xp,xfield)
    assert_allclose(yp,yfield)
    assert_allclose(zp,zfield)



from yt.fields.xray_emission_fields import \
    add_xray_emissivity_field
from yt.utilities.answer_testing.framework import \
    requires_ds, can_run_ds, data_dir_load, \
    ProjectionValuesTest, FieldValuesTest

def setup():
    from yt.config import ytcfg
    ytcfg["yt","__withintesting"] = "True"

def check_xray_fields(ds_fn, fields):
    if not can_run_ds(ds_fn): return
    dso = [ None, ("sphere", ("m", (0.1, 'unitary')))]
    for field in fields:
        for dobj_name in dso:
            for axis in [0, 1, 2]:
                yield ProjectionValuesTest(ds_fn, axis, field, 
                                           None, dobj_name)
            yield FieldValuesTest(ds_fn, field, dobj_name)

sloshing = "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0300"
@requires_ds(sloshing, big_data=True)
def test_sloshing_apec():
    ds = data_dir_load(sloshing)
    fields = add_xray_emissivity_field(ds, 0.5, 7.0, table_type="apec", 
                                       metallicity=0.3)
    for test in check_xray_fields(ds, fields):
        test_sloshing_apec.__name__ = test.description
        yield test

d9p = "D9p_500/10MpcBox_HartGal_csf_a0.500.d"
@requires_ds(d9p, big_data=True)
def test_d9p_cloudy():
    ds = data_dir_load(d9p)
    fields = add_xray_emissivity_field(ds, 0.5, 2.0, redshift=ds.current_redshift,
                                       table_type="cloudy", cosmology=ds.cosmology,
                                       metallicity=("gas", "metallicity"))
    for test in check_xray_fields(ds, fields):
        test_d9p_cloudy.__name__ = test.description
        yield test



import os
import shelve
import shutil
import tempfile
import unittest

import numpy as np
from nose.tools import assert_raises
from numpy.testing import assert_array_equal

from yt.data_objects.data_containers import YTDataContainer
from yt.data_objects.particle_filters import particle_filter
from yt.testing import assert_equal, fake_random_ds, fake_amr_ds,\
    fake_particle_ds, requires_module
from yt.utilities.exceptions import YTFieldNotFound, YTException

class TestDataContainers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_yt_data_container(self):
        # Test if ds could be None
        with assert_raises(RuntimeError) as err:
            YTDataContainer(None, None)
        desired = ('Error: ds must be set either through class'
                   ' type or parameter to the constructor')
        assert_equal(str(err.exception), desired)

        # Test if field_data key exists
        ds = fake_random_ds(5)
        proj = ds.proj("density", 0, data_source=ds.all_data())
        assert_equal('px' in proj.keys(), True)
        assert_equal('pz' in proj.keys(), False)

        # Delete the key and check if exits
        del proj['px']
        assert_equal('px' in proj.keys(), False)
        del proj['density']
        assert_equal('density' in proj.keys(), False)

        # Delete a non-existent field
        with assert_raises(YTFieldNotFound) as ex:
            del proj['p_mass']
        desired = "Could not find field '('stream', 'p_mass')' in UniformGridData."
        assert_equal(str(ex.exception), desired)

    def test_write_out(self):
        filename = "sphere.txt"
        ds = fake_particle_ds()
        sp = ds.sphere(ds.domain_center, 0.25)
        sp.write_out(filename, fields=["cell_volume"])

        with open(filename, "r") as file:
            file_row_1 = file.readline()
            file_row_2 = file.readline()
            file_row_2 = np.array(file_row_2.split('\t'), dtype=np.float64)
        sorted_keys = sorted(sp.field_data.keys())
        keys = [str(k) for k in sorted_keys]
        keys = "\t".join(["#"] + keys + ["\n"])
        data = [sp.field_data[k][0] for k in sorted_keys]

        assert_equal(keys, file_row_1)
        assert_array_equal(data, file_row_2)

        # Test for exception
        with assert_raises(YTException) as ex:
            sp.write_out(filename, fields=["particle_position_x"])
        desired = ("Field type ['all'] of the supplied field ['particle_position_x']"
                   " is in consistent with field type 'gas'.")
        assert_equal(str(ex.exception)[:50], desired[:50])

    def test_save_object(self):
        ds = fake_particle_ds()
        sp = ds.sphere(ds.domain_center, 0.25)
        sp.save_object("my_sphere_1", filename="test_save_obj")
        obj = shelve.open("test_save_obj", protocol=-1)
        loaded_sphere = obj["my_sphere_1"][1]
        obj.close()
        assert_array_equal(loaded_sphere.center, sp.center)
        assert_equal(loaded_sphere.radius, sp.radius)
        for k in loaded_sphere._key_fields:
            assert_array_equal(loaded_sphere[k], sp[k])

        # Object is saved but retrieval is not working
        # sp.save_object("my_sphere_2")
        # loaded_sphere = ds.index.load_object("my_sphere_2")
        # for k in loaded_sphere._key_fields:
        #     assert_array_equal(loaded_sphere[k], sp[k])

    @requires_module("pandas")
    def test_to_dataframe(self):
        fields = ["density", "velocity_z"]
        ds = fake_random_ds(6)
        dd = ds.all_data()
        df = dd.to_dataframe(fields)
        assert_array_equal(dd[fields[0]], df[fields[0]])
        assert_array_equal(dd[fields[1]], df[fields[1]])

    @requires_module("astropy")
    def test_to_astropy_table(self):
        from yt.units.yt_array import YTArray
        fields = ["density", "velocity_z"]
        ds = fake_random_ds(6)
        dd = ds.all_data()
        at1 = dd.to_astropy_table(fields)
        assert_array_equal(dd[fields[0]].d, at1[fields[0]].value)
        assert_array_equal(dd[fields[1]].d, at1[fields[1]].value)
        assert dd[fields[0]].units == YTArray.from_astropy(at1[fields[0]]).units
        assert dd[fields[1]].units == YTArray.from_astropy(at1[fields[1]]).units

    def test_std(self):
        ds = fake_random_ds(3)
        ds.all_data().std('density', weight="velocity_z")

    def test_to_frb(self):
        # Test cylindrical geometry
        fields = ["density", "cell_mass"]
        ds = fake_amr_ds(fields=fields, geometry="cylindrical", particles=16**3)
        dd = ds.all_data()
        proj = ds.proj("density", weight_field="cell_mass", axis=1, data_source=dd)
        frb = proj.to_frb((1.0, 'unitary'), 64)
        assert_equal(frb.radius, (1.0, 'unitary'))
        assert_equal(frb.buff_size, 64)

    def test_extract_isocontours(self):
        # Test isocontour properties for AMRGridData
        ds = fake_amr_ds(fields=["density", "cell_mass"], particles=16**3)
        dd = ds.all_data()
        q = dd.quantities["WeightedAverageQuantity"]
        rho = q("density", weight="cell_mass")
        dd.extract_isocontours("density", rho, "triangles.obj", True)
        dd.calculate_isocontour_flux("density", rho, "x", "y", "z", "dx")

        # Test error in case of ParticleData
        ds = fake_particle_ds()
        dd = ds.all_data()
        q = dd.quantities["WeightedAverageQuantity"]
        rho = q("particle_velocity_x", weight="particle_mass")
        with assert_raises(NotImplementedError):
            dd.extract_isocontours("density", rho, sample_values='x')

    def test_derived_field(self):
        # Test that derived field on filtered particles do not require
        # their parent field to be created
        ds = fake_particle_ds()
        dd = ds.all_data()

        @particle_filter(requires=['particle_mass'], filtered_type='io')
        def massive(pfilter, data):
            return data[(pfilter.filtered_type, 'particle_mass')].to('code_mass') > 0.5

        ds.add_particle_filter('massive')

        def fun(field, data):
            return data[field.name[0], 'particle_mass']

        # Add the field to the massive particles
        ds.add_field(('massive', 'test'), function=fun,
                     sampling_type='particle', units='code_mass')

        expected_size = (dd['io', 'particle_mass'].to('code_mass') > 0.5).sum()

        fields_to_test = (f for f in ds.derived_field_list
                          if f[0] == 'massive')

        def test_this(fname):
            data = dd[fname]
            assert_equal(data.shape[0], expected_size)

        for fname in fields_to_test:
            test_this(fname)



"""
This is a library of yt-defined exceptions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


# We don't need to import 'exceptions'
import os.path

class YTException(Exception):
    def __init__(self, message = None, ds = None):
        Exception.__init__(self, message)
        self.ds = ds

# Data access exceptions:

class YTOutputNotIdentified(YTException):
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return "Supplied %s %s, but could not load!" % (
            self.args, self.kwargs)

class YTSphereTooSmall(YTException):
    def __init__(self, ds, radius, smallest_cell):
        YTException.__init__(self, ds=ds)
        self.radius = radius
        self.smallest_cell = smallest_cell

    def __str__(self):
        return "%0.5e < %0.5e" % (self.radius, self.smallest_cell)

class YTAxesNotOrthogonalError(YTException):
    def __init__(self, axes):
        self.axes = axes

    def __str__(self):
        return "The supplied axes are not orthogonal.  %s" % (self.axes)

class YTNoDataInObjectError(YTException):
    def __init__(self, obj):
        self.obj_type = getattr(obj, "_type_name", "")

    def __str__(self):
        s = "The object requested has no data included in it."
        if self.obj_type == "slice":
            s += "  It may lie on a grid face.  Try offsetting slightly."
        return s

class YTFieldNotFound(YTException):
    def __init__(self, fname, ds):
        self.fname = fname
        self.ds = ds

    def __str__(self):
        return "Could not find field '%s' in %s." % (self.fname, self.ds)

class YTParticleTypeNotFound(YTException):
    def __init__(self, fname, ds):
        self.fname = fname
        self.ds = ds

    def __str__(self):
        return ("Could not find particle_type '%s' in %s." % (self.fname, self.ds))

class YTSceneFieldNotFound(YTException):
    pass

class YTCouldNotGenerateField(YTFieldNotFound):
    def __str__(self):
        return "Could field '%s' in %s could not be generated." % (self.fname, self.ds)

class YTFieldTypeNotFound(YTException):
    def __init__(self, ftype, ds=None):
        self.ftype = ftype
        self.ds = ds

    def __str__(self):
        if self.ds is not None and \
          self.ftype in self.ds.particle_types:
            return ("Could not find field type '%s'.  " +
                    "This field type is a known particle type for this dataset.  " +
                    "Try adding this field with particle_type=True.") % self.ftype
        else:
            return "Could not find field type '%s'." % (self.ftype)

class YTSimulationNotIdentified(YTException):
    def __init__(self, sim_type):
        YTException.__init__(self)
        self.sim_type = sim_type

    def __str__(self):
        return "Simulation time-series type %s not defined." % self.sim_type

class YTCannotParseFieldDisplayName(YTException):
    def __init__(self, field_name, display_name, mathtext_error):
        self.field_name = field_name
        self.display_name = display_name
        self.mathtext_error = mathtext_error

    def __str__(self):
        return ("The display name \"%s\" "
                "of the derived field %s "
                "contains the following LaTeX parser errors:\n" ) \
                % (self.display_name, self.field_name) + self.mathtext_error

class YTCannotParseUnitDisplayName(YTException):
    def __init__(self, field_name, unit_name, mathtext_error):
        self.field_name = field_name
        self.unit_name = unit_name
        self.mathtext_error = mathtext_error

    def __str__(self):
        return ("The unit display name \"%s\" "
                "of the derived field %s "
                "contains the following LaTeX parser errors:\n" ) \
            % (self.unit_name, self.field_name) + self.mathtext_error

class InvalidSimulationTimeSeries(YTException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class MissingParameter(YTException):
    def __init__(self, ds, parameter):
        YTException.__init__(self, ds=ds)
        self.parameter = parameter

    def __str__(self):
        return "dataset %s is missing %s parameter." % \
            (self.ds, self.parameter)

class NoStoppingCondition(YTException):
    def __init__(self, ds):
        YTException.__init__(self, ds=ds)

    def __str__(self):
        return "Simulation %s has no stopping condition.  StopTime or StopCycle should be set." % \
            self.ds

class YTNotInsideNotebook(YTException):
    def __str__(self):
        return "This function only works from within an IPython Notebook."

class YTGeometryNotSupported(YTException):
    def __init__(self, geom):
        self.geom = geom

    def __str__(self):
        return "We don't currently support %s geometry" % self.geom

class YTCoordinateNotImplemented(YTException):
    def __str__(self):
        return "This coordinate is not implemented for this geometry type."

class YTUnitNotRecognized(YTException):
    def __init__(self, unit):
        self.unit = unit

    def __str__(self):
        return "This dataset doesn't recognize %s" % self.unit

class YTUnitOperationError(YTException, ValueError):
    def __init__(self, operation, unit1, unit2=None):
        self.operation = operation
        self.unit1 = unit1
        self.unit2 = unit2
        YTException.__init__(self)

    def __str__(self):
        err = "The %s operator for YTArrays with units (%s) " % (self.operation, self.unit1, )
        if self.unit2 is not None:
            err += "and (%s) " % self.unit2
        err += "is not well defined."
        return err

class YTUnitConversionError(YTException):
    def __init__(self, unit1, dimension1, unit2, dimension2):
        self.unit1 = unit1
        self.unit2 = unit2
        self.dimension1 = dimension1
        self.dimension2 = dimension2
        YTException.__init__(self)

    def __str__(self):
        err = "Unit dimensionalities do not match. Tried to convert between " \
          "%s (dim %s) and %s (dim %s)." \
          % (self.unit1, self.dimension1, self.unit2, self.dimension2)
        return err

class YTUnitsNotReducible(YTException):
    def __init__(self, unit, units_base):
        self.unit = unit
        self.units_base = units_base
        YTException.__init__(self)

    def __str__(self):
        err = "The unit '%s' cannot be reduced to a single expression within " \
          "the %s base system of units." % (self.unit, self.units_base)
        return err

class YTEquivalentDimsError(YTUnitOperationError):
    def __init__(self, old_units, new_units, base):
        self.old_units = old_units
        self.new_units = new_units
        self.base = base

    def __str__(self):
        err = "It looks like you're trying to convert between '%s' and '%s'. Try " \
          "using \"to_equivalent('%s', '%s')\" instead." % (self.old_units, self.new_units,
                                                            self.new_units, self.base)
        return err

class YTUfuncUnitError(YTException):
    def __init__(self, ufunc, unit1, unit2):
        self.ufunc = ufunc
        self.unit1 = unit1
        self.unit2 = unit2
        YTException.__init__(self)

    def __str__(self):
        err = "The NumPy %s operation is only allowed on objects with " \
              "identical units. Convert one of the arrays to the other\'s " \
              "units first. Received units (%s) and (%s)." % \
              (self.ufunc, self.unit1, self.unit2)
        return err

class YTIterableUnitCoercionError(YTException):
    def __init__(self, quantity_list):
        self.quantity_list = quantity_list

    def __str__(self):
        err = "Received a list or tuple of quantities with nonuniform units: " \
              "%s" % self.quantity_list
        return err

class YTFieldUnitError(YTException):
    def __init__(self, field_info, returned_units):
        self.msg = ("The field function associated with the field '%s' returned "
                    "data with units '%s' but was defined with units '%s'.")
        self.msg = self.msg % (field_info.name, returned_units, field_info.units)

    def __str__(self):
        return self.msg

class YTFieldUnitParseError(YTException):
    def __init__(self, field_info):
        self.msg = ("The field '%s' has unparseable units '%s'.")
        self.msg = self.msg % (field_info.name, field_info.units)

    def __str__(self):
        return self.msg

class YTSpatialFieldUnitError(YTException):
    def __init__(self, field):
        msg = ("Field '%s' is a spatial field but has unknown units but "
               "spatial fields must have explicitly defined units. Add the "
               "field with explicit 'units' to clear this error.")
        self.msg = msg % (field,)

    def __str__(self):
        return self.msg

class YTHubRegisterError(YTException):
    def __str__(self):
        return "You must create an API key before uploading.  See " + \
               "https://data.yt-project.org/getting_started.html"

class YTNoFilenamesMatchPattern(YTException):
    def __init__(self, pattern):
        self.pattern = pattern

    def __str__(self):
        return "No filenames were found to match the pattern: " + \
               "'%s'" % (self.pattern)

class YTNoOldAnswer(YTException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "There is no old answer available.\n" + \
               str(self.path)

class YTNoAnswerNameSpecified(YTException):
    def __init__(self, message=None):
        if message is None or message == "":
            message = ("Answer name not provided for the answer testing test."
                       "\n  Please specify --answer-name=<answer_name> in"
                       " command line mode or in AnswerTestingTest.answer_name"
                       " variable."
                       )
        self.message = message

    def __str__(self):
        return str(self.message)

class YTCloudError(YTException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "Failed to retrieve cloud data. Connection may be broken.\n" + \
               str(self.path)

class YTEllipsoidOrdering(YTException):
    def __init__(self, ds, A, B, C):
        YTException.__init__(self, ds=ds)
        self._A = A
        self._B = B
        self._C = C

    def __str__(self):
        return "Must have A>=B>=C"

class EnzoTestOutputFileNonExistent(YTException):
    def __init__(self, filename):
        self.filename = filename
        self.testname = os.path.basename(os.path.dirname(filename))

    def __str__(self):
        return "Enzo test output file (OutputLog) not generated for: " + \
            "'%s'" % (self.testname) + ".\nTest did not complete."

class YTNoAPIKey(YTException):
    def __init__(self, service, config_name):
        self.service = service
        self.config_name = config_name

    def __str__(self):
        return "You need to set an API key for %s in ~/.config/yt/ytrc as %s" % (
            self.service, self.config_name)

class YTTooManyVertices(YTException):
    def __init__(self, nv, fn):
        self.nv = nv
        self.fn = fn

    def __str__(self):
        s = "There are too many vertices (%s) to upload to Sketchfab. " % (self.nv)
        s += "Your model has been saved as %s .  You should upload manually." % (self.fn)
        return s

class YTInvalidWidthError(YTException):
    def __init__(self, width):
        self.error = "width (%s) is invalid" % str(width)

    def __str__(self):
        return str(self.error)

class YTFieldNotParseable(YTException):
    def __init__(self, field):
        self.field = field

    def __str__(self):
        return "Cannot identify field %s" % (self.field,)

class YTDataSelectorNotImplemented(YTException):
    def __init__(self, class_name):
        self.class_name = class_name

    def __str__(self):
        return "Data selector '%s' not implemented." % (self.class_name)

class YTParticleDepositionNotImplemented(YTException):
    def __init__(self, class_name):
        self.class_name = class_name

    def __str__(self):
        return "Particle deposition method '%s' not implemented." % (self.class_name)

class YTDomainOverflow(YTException):
    def __init__(self, mi, ma, dle, dre):
        self.mi = mi
        self.ma = ma
        self.dle = dle
        self.dre = dre

    def __str__(self):
        return "Particle bounds %s and %s exceed domain bounds %s and %s" % (
            self.mi, self.ma, self.dle, self.dre)

class YTIntDomainOverflow(YTException):
    def __init__(self, dims, dd):
        self.dims = dims
        self.dd = dd

    def __str__(self):
        return "Integer domain overflow: %s in %s" % (
            self.dims, self.dd)

class YTIllDefinedFilter(YTException):
    def __init__(self, filter, s1, s2):
        self.filter = filter
        self.s1 = s1
        self.s2 = s2

    def __str__(self):
        return "Filter '%s' ill-defined.  Applied to shape %s but is shape %s." % (
            self.filter, self.s1, self.s2)

class YTIllDefinedParticleFilter(YTException):
    def __init__(self, filter, missing):
        self.filter = filter
        self.missing = missing

    def __str__(self):
        msg = ("\nThe fields\n\t{},\nrequired by the \"{}\" particle filter, "
               "are not defined for this dataset.")
        f = self.filter
        return msg.format("\n".join([str(m) for m in self.missing]), f.name)

class YTIllDefinedBounds(YTException):
    def __init__(self, lb, ub):
        self.lb = lb
        self.ub = ub

    def __str__(self):
        v =  "The bounds %0.3e and %0.3e are ill-defined. " % (self.lb, self.ub)
        v += "Typically this happens when a log binning is specified "
        v += "and zero or negative values are given for the bounds."
        return v

class YTObjectNotImplemented(YTException):
    def __init__(self, ds, obj_name):
        self.ds = ds
        self.obj_name = obj_name

    def __str__(self):
        v  = r"The object type '%s' is not implemented for the dataset "
        v += r"'%s'."
        return v % (self.obj_name, self.ds)

class YTParticleOutputFormatNotImplemented(YTException):
    def __str__(self):
        return "The particle output format is not supported."

class YTFileNotParseable(YTException):
    def __init__(self, fname, line):
        self.fname = fname
        self.line = line

    def __str__(self):
        v = r"Error while parsing file %s at line %s"
        return v % (self.fname, self.line)

class YTRockstarMultiMassNotSupported(YTException):
    def __init__(self, mi, ma, ptype):
        self.mi = mi
        self.ma = ma
        self.ptype = ptype

    def __str__(self):
        v = "Particle type '%s' has minimum mass %0.3e and maximum " % (
            self.ptype, self.mi)
        v += "mass %0.3e.  Multi-mass particles are not currently supported." % (
            self.ma)
        return v

class YTTooParallel(YTException):
    def __str__(self):
        return "You've used too many processors for this dataset."


class YTElementTypeNotRecognized(YTException):
    def __init__(self, dim, num_nodes):
        self.dim = dim
        self.num_nodes = num_nodes

    def __str__(self):
        return "Element type not recognized - dim = %s, num_nodes = %s" % (
            self.dim, self.num_nodes)


class YTDuplicateFieldInProfile(Exception):
    def __init__(self, field, new_spec, old_spec):
        self.field = field
        self.new_spec = new_spec
        self.old_spec = old_spec

    def __str__(self):
        r = """Field %s already exists with field spec:
               %s
               But being asked to add it with:
               %s""" % (self.field, self.old_spec, self.new_spec)
        return r

class YTInvalidPositionArray(Exception):
    def __init__(self, shape, dimensions):
        self.shape = shape
        self.dimensions = dimensions

    def __str__(self):
        r = """Position arrays must be length and shape (N,3).
               But this one has %s and %s.""" % (self.dimensions, self.shape)
        return r

class YTIllDefinedCutRegion(Exception):
    def __init__(self, conditions):
        self.conditions = conditions

    def __str__(self):
        r = """Can't mix particle/discrete and fluid/mesh conditions or
               quantities.  Conditions specified:
            """
        r += "\n".join([c for c in self.conditions])
        return r

class YTMixedCutRegion(Exception):
    def __init__(self, conditions, field):
        self.conditions = conditions
        self.field = field

    def __str__(self):
        r = """Can't mix particle/discrete and fluid/mesh conditions or
               quantities.  Field: %s and Conditions specified:
            """ % (self.field,)
        r += "\n".join([c for c in self.conditions])
        return r

class YTGDFAlreadyExists(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return "A file already exists at %s and overwrite=False." % self.filename

class YTGDFUnknownGeometry(Exception):
    def __init__(self, geometry):
        self.geometry = geometry

    def __str__(self):
        return '''Unknown geometry %i. Please refer to GDF standard
                  for more information''' % self.geometry

class YTInvalidUnitEquivalence(Exception):
    def __init__(self, equiv, unit1, unit2):
        self.equiv = equiv
        self.unit1 = unit1
        self.unit2 = unit2

    def __str__(self):
        return "The unit equivalence '%s' does not exist for the units '%s' and '%s'." % (self.equiv,
                                                                                          self.unit1,
                                                                                          self.unit2)

class YTPlotCallbackError(Exception):
    def __init__(self, callback, error):
        self.callback = 'annotate_' + callback
        self.error = error

    def __str__(self):
        msg = '%s callback failed with the following error: %s'
        return msg % (self.callback, self.error)

class YTPixelizeError(YTException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class YTDimensionalityError(YTException):
    def __init__(self, wrong, right):
        self.wrong = wrong
        self.right = right

    def __str__(self):
        return 'Dimensionality specified was %s but we need %s' % (
            self.wrong, self.right)

class YTInvalidShaderType(YTException):
    def __init__(self, source):
        self.source = source

    def __str__(self):
        return "Can't identify shader_type for file '%s.'" % (self.source)

class YTInvalidFieldType(YTException):
    def __init__(self, fields):
        self.fields = fields

    def __str__(self):
        msg = ("\nSlicePlot, ProjectionPlot, and OffAxisProjectionPlot can only "
               "plot fields that\n"
               "are defined on a mesh, but received the following particle "
               "fields:\n\n"
               "    %s\n\n"
               "Did you mean to use ParticlePlot or plot a deposited particle "
               "field instead?" % self.fields)
        return msg

class YTUnknownUniformKind(YTException):
    def __init__(self, kind):
        self.kind = kind

    def __str__(self):
        return "Can't determine kind specification for %s" % (self.kind)

class YTUnknownUniformSize(YTException):
    def __init__(self, size_spec):
        self.size_spec = size_spec

    def __str__(self):
        return "Can't determine size specification for %s" % (self.size_spec)

class YTDataTypeUnsupported(YTException):
    def __init__(self, this, supported):
        self.supported = supported
        self.this = this

    def __str__(self):
        v = "This operation is not supported for data of geometry %s; " % self.this
        v += "It supports data of geometries %s" % (self.supported,)
        return v

class YTBoundsDefinitionError(YTException):
    def __init__(self, message, bounds):
        self.bounds = bounds
        self.message = message

    def __str__(self):
        v  = "This operation has encountered a bounds error: "
        v += self.message
        v += " Specified bounds are '%s'." % (self.bounds,)
        return v

def screen_one_element_list(lis):
    if len(lis) == 1:
        return lis[0]
    return lis

class YTIllDefinedProfile(YTException):
    def __init__(self, bin_fields, fields, weight_field, is_pfield):
        nbin = len(bin_fields)
        nfields = len(fields)
        self.bin_fields = screen_one_element_list(bin_fields)
        self.bin_fields_ptype = screen_one_element_list(is_pfield[:nbin])
        self.fields = screen_one_element_list(fields)
        self.fields_ptype = screen_one_element_list(is_pfield[nbin:nbin+nfields])
        self.weight_field = weight_field
        if self.weight_field is not None:
            self.weight_field_ptype = is_pfield[-1]

    def __str__(self):
        msg = (
            "\nCannot create a profile object that mixes particle and mesh "
            "fields.\n\n"
            "Received the following bin_fields:\n\n"
            "   %s, particle_type = %s\n\n"
            "Profile fields:\n\n"
            "   %s, particle_type = %s\n"
        )
        msg = msg % (
            self.bin_fields, self.bin_fields_ptype,
            self.fields, self.fields_ptype
        )

        if self.weight_field is not None:
            weight_msg = "\nAnd weight field:\n\n   %s, particle_type = %s\n"
            weight_msg = weight_msg % (
                self.weight_field, self.weight_field_ptype)
        else:
            weight_msg = ""

        return msg + weight_msg

class YTProfileDataShape(YTException):
    def __init__(self, field1, shape1, field2, shape2):
        self.field1 = field1
        self.shape1 = shape1
        self.field2 = field2
        self.shape2 = shape2

    def __str__(self):
        return ("Profile fields must have same shape: %s has " +
                "shape %s and %s has shape %s.") % \
                (self.field1, self.shape1, self.field2, self.shape2)

class YTBooleanObjectError(YTException):
    def __init__(self, bad_object):
        self.bad_object = bad_object

    def __str__(self):
        v  = "Supplied:\n%s\nto a boolean operation" % (self.bad_object)
        v += " but it is not a YTSelectionContainer3D object."
        return v

class YTBooleanObjectsWrongDataset(YTException):
    def __init__(self):
        pass

    def __str__(self):
        return "Boolean data objects must share a common dataset object."

class YTIllDefinedAMR(YTException):
    def __init__(self, level, axis):
        self.level = level
        self.axis = axis

    def __str__(self):
        msg = (
            "Grids on the level {} are not properly aligned with cell edges "
            "on the parent level ({} axis)"
        ).format(self.level, self.axis)
        return msg

class YTIllDefinedParticleData(YTException):
    pass

class YTIllDefinedAMRData(YTException):
    pass

class YTInconsistentGridFieldShape(YTException):
    def __init__(self, shapes):
        self.shapes = shapes

    def __str__(self):
        msg = "Not all grid-based fields have the same shape!\n"
        for name, shape in self.shapes:
            msg += "    Field {} has shape {}.\n".format(name, shape)
        return msg

class YTInconsistentParticleFieldShape(YTException):
    def __init__(self, ptype, shapes):
        self.ptype = ptype
        self.shapes = shapes

    def __str__(self):
        msg = (
            "Not all fields with field type '{}' have the same shape!\n"
        ).format(self.ptype)
        for name, shape in self.shapes:
            field = (self.ptype, name)
            msg += "    Field {} has shape {}.\n".format(field, shape)
        return msg

class YTInconsistentGridFieldShapeGridDims(YTException):
    def __init__(self, shapes, grid_dims):
        self.shapes = shapes
        self.grid_dims = grid_dims

    def __str__(self):
        msg = "Not all grid-based fields match the grid dimensions! "
        msg += "Grid dims are {}, ".format(self.grid_dims)
        msg += "and the following fields have shapes that do not match them:\n"
        for name, shape in self.shapes:
            if shape != self.grid_dims:
                msg += "    Field {} has shape {}.\n".format(name, shape)
        return msg

class YTCommandRequiresModule(YTException):
    def __init__(self, module):
        self.module = module

    def __str__(self):
        msg = "This command requires \"%s\" to be installed.\n\n" % self.module
        msg += "Please install \"%s\" with the package manager " % self.module
        msg += "appropriate for your python environment, e.g.:\n"
        msg += "  conda install %s\n" % self.module
        msg += "or:\n"
        msg += "  pip install %s\n" % self.module
        return msg

class YTArrayTooLargeToDisplay(YTException):
    def __init__(self, size, max_size):
        self.size = size
        self.max_size = max_size

    def __str__(self):
        msg  = "The requested array is of size %s.\n" % self.size
        msg += "We do not support displaying arrays larger\n"
        msg += "than size %s." % self.max_size
        return msg



"""
API for AdaptaHOP frontend




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
     AdaptaHOPDataset

from .io import \
     IOHandlerAdaptaHOPBinary

from .fields import \
     AdaptaHOPFieldInfo

from . import tests



"""
openPMD-specific fields



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
# Copyright (c) 2015, Daniel Grassinger (HZDR)
# Copyright (c) 2016, Fabian Koller (HZDR)
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import numpy as np

from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.magnetic_field import setup_magnetic_field_aliases
from yt.frontends.open_pmd.misc import \
    parse_unit_dimension, \
    is_const_component
from yt.units.yt_array import YTQuantity
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5
from yt.utilities.physical_constants import \
    speed_of_light, \
    mu_0


def setup_poynting_vector(self):
    def _get_poyn(axis):
        def poynting(field, data):
            u = mu_0**-1
            if axis in "x":
                return u * (data["E_y"] * data["magnetic_field_z"] - data["E_z"] * data["magnetic_field_y"])
            elif axis in "y":
                return u * (data["E_z"] * data["magnetic_field_x"] - data["E_x"] * data["magnetic_field_z"])
            elif axis in "z":
                return u * (data["E_x"] * data["magnetic_field_y"] - data["E_y"] * data["magnetic_field_x"])

        return poynting

    for ax in "xyz":
        self.add_field(("openPMD", "poynting_vector_%s" % ax), sampling_type="cell",
                       function=_get_poyn(ax),
                       units="W/m**2")


def setup_kinetic_energy(self, ptype):
    def _kin_en(field, data):
        p2 = (data[ptype, "particle_momentum_x"] ** 2 +
              data[ptype, "particle_momentum_y"] ** 2 +
              data[ptype, "particle_momentum_z"] ** 2)
        mass = data[ptype, "particle_mass"] * data[ptype, "particle_weighting"]
        return speed_of_light * np.sqrt(p2 + mass ** 2 * speed_of_light ** 2) - mass * speed_of_light ** 2

    self.add_field((ptype, "particle_kinetic_energy"), sampling_type="particle",
                   function=_kin_en,
                   units="kg*m**2/s**2")

def setup_velocity(self, ptype):
    def _get_vel(axis):
        def velocity(field, data):
            c = speed_of_light
            momentum = data[ptype, "particle_momentum_{}".format(axis)]
            mass = data[ptype, "particle_mass"]
            weighting = data[ptype, "particle_weighting"]
            return momentum / np.sqrt(
                (mass * weighting) ** 2 +
                (momentum ** 2) / (c ** 2)
            )

        return velocity

    for ax in "xyz":
        self.add_field((ptype, "particle_velocity_%s" % ax), sampling_type="particle",
                       function=_get_vel(ax),
                       units="m/s")


def setup_absolute_positions(self, ptype):
    def _abs_pos(axis):
        def ap(field, data):
            return np.add(data[ptype, "particle_positionCoarse_{}".format(axis)],
                          data[ptype, "particle_positionOffset_{}".format(axis)])

        return ap

    for ax in "xyz":
        self.add_field((ptype, "particle_position_%s" % ax), sampling_type="particle",
                       function=_abs_pos(ax),
                       units="m")


class OpenPMDFieldInfo(FieldInfoContainer):
    """Specifies which fields from the dataset yt should know about.

    ``self.known_other_fields`` and ``self.known_particle_fields`` must be populated.
    Entries for both of these lists must be tuples of the form
        ("name", ("units", ["fields", "to", "alias"], "display_name"))
    These fields will be represented and handled in yt in the way you define them here.
    The fields defined in both ``self.known_other_fields`` and ``self.known_particle_fields`` will only be added
    to a dataset (with units, aliases, etc), if they match any entry in the ``OpenPMDHierarchy``'s ``self.field_list``.

    Notes
    -----

    Contrary to many other frontends, we dynamically obtain the known fields from the simulation output.
    The openPMD markup is extremely flexible - names, dimensions and the number of individual datasets
    can (and very likely will) vary.

    openPMD states that names of records and their components are only allowed to contain the
        characters a-Z,
        the numbers 0-9
        and the underscore _
        (equivalently, the regex \w).
    Since yt widely uses the underscore in field names, openPMD's underscores (_) are replaced by hyphen (-).

    Derived fields will automatically be set up, if names and units of your known on-disk (or manually derived)
    fields match the ones in [1].

    References
    ----------
    * http://yt-project.org/docs/dev/analyzing/fields.html
    * http://yt-project.org/docs/dev/developing/creating_frontend.html#data-meaning-structures
    * https://github.com/openPMD/openPMD-standard/blob/latest/STANDARD.md
    * [1] http://yt-project.org/docs/dev/reference/field_list.html#universal-fields
    """
    _mag_fields = []

    def __init__(self, ds, field_list):
        f = ds._handle
        bp = ds.base_path
        mp = ds.meshes_path
        pp = ds.particles_path

        try:
            fields = f[bp + mp]
            for fname in fields.keys():
                field = fields[fname]
                if type(field) is h5.Dataset or is_const_component(field):
                    # Don't consider axes. This appears to be a vector field of single dimensionality
                    ytname = str("_".join([fname.replace("_", "-")]))
                    parsed = parse_unit_dimension(np.asarray(field.attrs["unitDimension"], dtype=np.int))
                    unit = str(YTQuantity(1, parsed).units)
                    aliases = []
                    # Save a list of magnetic fields for aliasing later on
                    # We can not reasonably infer field type/unit by name in openPMD
                    if unit == "T" or unit == "kg/(A*s**2)":
                        self._mag_fields.append(ytname)
                    self.known_other_fields += ((ytname, (unit, aliases, None)),)
                else:
                    for axis in field.keys():
                        ytname = str("_".join([fname.replace("_", "-"), axis]))
                        parsed = parse_unit_dimension(np.asarray(field.attrs["unitDimension"], dtype=np.int))
                        unit = str(YTQuantity(1, parsed).units)
                        aliases = []
                        # Save a list of magnetic fields for aliasing later on
                        # We can not reasonably infer field type by name in openPMD
                        if unit == "T" or unit == "kg/(A*s**2)":
                            self._mag_fields.append(ytname)
                        self.known_other_fields += ((ytname, (unit, aliases, None)),)
            for i in self.known_other_fields:
                mylog.debug("open_pmd - known_other_fields - {}".format(i))
        except(KeyError, TypeError, AttributeError):
            pass

        try:
            particles = f[bp + pp]
            for pname in particles.keys():
                species = particles[pname]
                for recname in species.keys():
                    try:
                        record = species[recname]
                        parsed = parse_unit_dimension(record.attrs["unitDimension"])
                        unit = str(YTQuantity(1, parsed).units)
                        ytattrib = str(recname).replace("_", "-")
                        if ytattrib == "position":
                            # Symbolically rename position to preserve yt's interpretation of the pfield
                            # particle_position is later derived in setup_absolute_positions in the way yt expects it
                            ytattrib = "positionCoarse"
                        if type(record) is h5.Dataset or is_const_component(record):
                            name = ["particle", ytattrib]
                            self.known_particle_fields += ((str("_".join(name)), (unit, [], None)),)
                        else:
                            for axis in record.keys():
                                aliases = []
                                name = ["particle", ytattrib, axis]
                                ytname = str("_".join(name))
                                self.known_particle_fields += ((ytname, (unit, aliases, None)),)
                    except(KeyError):
                        if recname != "particlePatches":
                            mylog.info("open_pmd - {}_{} does not seem to have unitDimension".format(pname, recname))
            for i in self.known_particle_fields:
                mylog.debug("open_pmd - known_particle_fields - {}".format(i))
        except(KeyError, TypeError, AttributeError):
            pass

        super(OpenPMDFieldInfo, self).__init__(ds, field_list)

    def setup_fluid_fields(self):
        """Defines which derived mesh fields to create.

        If a field can not be calculated, it will simply be skipped.
        """
        # Set up aliases first so the setup for poynting can use them
        if len(self._mag_fields) > 0:
            setup_magnetic_field_aliases(self, "openPMD", self._mag_fields)
            setup_poynting_vector(self)

    def setup_particle_fields(self, ptype):
        """Defines which derived particle fields to create.

        This will be called for every entry in `OpenPMDDataset``'s ``self.particle_types``.
        If a field can not be calculated, it will simply be skipped.
        """
        setup_absolute_positions(self, ptype)
        setup_kinetic_energy(self, ptype)
        setup_velocity(self, ptype)
        super(OpenPMDFieldInfo, self).setup_particle_fields(ptype)



from yt.testing import \
    fake_amr_ds, _geom_transforms
from yt.utilities.answer_testing.framework import \
    AxialPixelizationTest

def test_axial_pixelization():
    for geom in sorted(_geom_transforms):
        ds = fake_amr_ds(geometry=geom)
        yield AxialPixelizationTest(ds)



"""
Unit test the sunyaev_zeldovich analysis module.
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.frontends.stream.api import load_uniform_grid
from yt.funcs import get_pbar
from yt.utilities.physical_ratios import \
    cm_per_kpc, \
    K_per_keV, \
    cm_per_km
from yt.utilities.physical_constants import \
    mh, \
    kboltz, \
    Tcmb, \
    hcgs, \
    clight, \
    sigma_thompson
from yt.testing import requires_module, assert_almost_equal
from yt.utilities.answer_testing.framework import requires_ds, \
    GenericArrayTest, data_dir_load, GenericImageTest
try:
    from yt.analysis_modules.sunyaev_zeldovich.projection import SZProjection, I0
except ImportError:
    pass
import numpy as np
try:
    import SZpack
except ImportError:
    pass

mue = 1./0.88
freqs = np.array([30., 90., 240.])

def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"

def full_szpack3d(ds, xo):
    data = ds.index.grids[0]
    dz = ds.index.get_smallest_dx().in_units("cm")
    nx,ny,nz = data["density"].shape
    dn = np.zeros((nx,ny,nz))
    Dtau = np.array(sigma_thompson*data["density"]/(mh*mue)*dz)
    Te = data["kT"].ndarray_view()
    betac = np.array(data["velocity_z"]/clight)
    pbar = get_pbar("Computing 3-D cell-by-cell S-Z signal for comparison.", nx)
    for i in range(nx):
        pbar.update(i)
        for j in range(ny):
            for k in range(nz):
                dn[i,j,k] = SZpack.compute_3d(xo, Dtau[i,j,k],
                                              Te[i,j,k], betac[i,j,k],
                                              1.0, 0.0, 0.0, 1.0e-5)
    pbar.finish()
    return np.array(I0*xo**3*np.sum(dn, axis=2))

def setup_cluster():

    R = 1000.
    r_c = 100.
    rho_c = 1.673e-26
    beta = 1.
    T0 = 4.
    nx,ny,nz = 16,16,16
    c = 0.17
    a_c = 30.
    a = 200.
    v0 = 300.*cm_per_km
    ddims = (nx,ny,nz)

    x, y, z = np.mgrid[-R:R:nx*1j,
                       -R:R:ny*1j,
                       -R:R:nz*1j]

    r = np.sqrt(x**2+y**2+z**2)

    dens = np.zeros(ddims)
    dens = rho_c*(1.+(r/r_c)**2)**(-1.5*beta)
    temp = T0*K_per_keV/(1.+r/a)*(c+r/a_c)/(1.+r/a_c)
    velz = v0*temp/(T0*K_per_keV)

    data = {}
    data["density"] = (dens, "g/cm**3")
    data["temperature"] = (temp, "K")
    data["velocity_x"] = (np.zeros(ddims), "cm/s")
    data["velocity_y"] = (np.zeros(ddims), "cm/s")
    data["velocity_z"] = (velz, "cm/s")

    L = 2 * R * cm_per_kpc
    bbox = np.array([[-0.5,0.5],[-0.5,0.5],[-0.5,0.5]]) * L

    ds = load_uniform_grid(data, ddims, length_unit='cm', bbox=bbox)
    ds.index

    return ds

@requires_module("SZpack")
def test_projection():
    ds = setup_cluster()
    nx,ny,nz = ds.domain_dimensions
    xinit = np.array(1.0e9*hcgs*freqs/(kboltz*Tcmb))
    szprj = SZProjection(ds, freqs, mue=mue, high_order=True)
    szprj.on_axis(2, nx=nx)
    deltaI = np.zeros((3,nx,ny))
    for i in range(3):
        deltaI[i,:,:] = full_szpack3d(ds, xinit[i])
        assert_almost_equal(
            deltaI[i,:,:], np.array(szprj["%d_GHz" % int(freqs[i])]), 6)

M7 = "DD0010/moving7_0010"
@requires_module("SZpack")
@requires_ds(M7)
def test_M7_onaxis():
    ds = data_dir_load(M7)
    szprj = SZProjection(ds, freqs)
    szprj.on_axis(2, nx=100)
    def onaxis_array_func():
        return szprj.data
    def onaxis_image_func(filename_prefix):
        szprj.write_png(filename_prefix)
    for test in [GenericArrayTest(ds, onaxis_array_func),
                 GenericImageTest(ds, onaxis_image_func, 12)]:
        test_M7_onaxis.__name__ = test.description
        yield test

@requires_module("SZpack")
@requires_ds(M7)
def test_M7_offaxis():
    ds = data_dir_load(M7)
    szprj = SZProjection(ds, freqs)
    szprj.off_axis(np.array([0.1,-0.2,0.4]), nx=100)
    def offaxis_array_func():
        return szprj.data
    def offaxis_image_func(filename_prefix):
        szprj.write_png(filename_prefix)
    for test in [GenericArrayTest(ds, offaxis_array_func),
                 GenericImageTest(ds, offaxis_image_func, 12)]:
        test_M7_offaxis.__name__ = test.description
        yield test



"""
API for ytData frontend




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .data_structures import \
    YTDataContainerDataset, \
    YTSpatialPlotDataset, \
    YTGridDataset, \
    YTGridHierarchy, \
    YTGrid, \
    YTNonspatialDataset, \
    YTNonspatialHierarchy, \
    YTNonspatialGrid, \
    YTProfileDataset, \
    YTClumpContainer, \
    YTClumpTreeDataset

from .io import \
    IOHandlerYTDataContainerHDF5, \
    IOHandlerYTGridHDF5, \
    IOHandlerYTSpatialPlotHDF5, \
    IOHandlerYTNonspatialhdf5

from .fields import \
    YTDataContainerFieldInfo, \
    YTGridFieldInfo

from .utilities import \
    save_as_dataset

from . import tests



"""
General field-related functions.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.utilities.lib.misc_utilities import \
    obtain_position_vector

def get_radius(data, field_prefix, ftype):
    unit_system = data.ds.unit_system
    center = data.get_field_parameter("center").in_base(unit_system.name)
    DW = (data.ds.domain_right_edge - data.ds.domain_left_edge).in_base(unit_system.name)
    # This is in cm**2 so it can be the destination for our r later.
    radius2 = data.ds.arr(np.zeros(data[ftype, field_prefix+"x"].shape,
                          dtype='float64'), 'cm**2')
    r = radius2.copy()
    if any(data.ds.periodicity):
        rdw = radius2.copy()
    for i, ax in enumerate('xyz'):
        # This will coerce the units, so we don't need to worry that we copied
        # it from a cm**2 array.
        np.subtract(data[ftype, "%s%s" % (field_prefix, ax)].in_base(unit_system.name),
                    center[i], r)
        if data.ds.periodicity[i] is True:
            np.abs(r, r)
            np.subtract(r, DW[i], rdw)
            np.abs(rdw, rdw)
            np.minimum(r, rdw, r)
        np.multiply(r, r, r)
        np.add(radius2, r, radius2)
        if data.ds.dimensionality < i+1:
            break
    # Now it's cm.
    np.sqrt(radius2, radius2)
    # Alias it, just for clarity.
    radius = radius2
    return radius

def get_periodic_rvec(data):
    coords = obtain_position_vector(data).d
    if sum(data.ds.periodicity) == 0: return coords
    le = data.ds.domain_left_edge.in_units("code_length").d
    dw = data.ds.domain_width.in_units("code_length").d
    for i in range(coords.shape[0]):
        if not data.ds.periodicity[i]: continue
        coords[i, ...] -= le[i]
        #figure out which measure is less
        mins = np.argmin([np.abs(np.mod(coords[i, ...],  dw[i])),
                         np.abs(np.mod(coords[i, ...], -dw[i]))],
                         axis=0)
        temp_coords = np.mod(coords[i, ...], dw[i])

        #Where second measure is better, updating temporary coords
        ii = mins==1
        temp_coords[ii] = np.mod(coords[i, ...], -dw[i])[ii]

        # Putting the temporary coords into the actual storage
        coords[i, ...] = temp_coords

        coords[i, ...] + le[i]

    return coords



"""
API for halo_mass_function



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.funcs import issue_deprecation_warning

issue_deprecation_warning(
    "The halo_mass_function module does not function correctly and has been "
    "deprecated. This code has been moved to the yt attic "
    "(https://github.com/yt-project/yt_attic) and will be removed in a "
    "future release.")

from .halo_mass_function import \
    HaloMassFcn, \
    TransferFunction, \
    integrate_inf



"""
HaloCatalog data-file handling function




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np

from yt.utilities.exceptions import YTDomainOverflow
from yt.funcs import \
    mylog, \
    parse_h5_attr

from yt.utilities.io_handler import \
    BaseIOHandler

from yt.utilities.lib.geometry_utils import compute_morton


class IOHandlerHaloCatalogHDF5(BaseIOHandler):
    _dataset_type = "halocatalog_hdf5"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        chunks = list(chunks)
        data_files = set([])
        # Only support halo reading for now.
        assert(len(ptf) == 1)
        assert(list(ptf.keys())[0] == "halos")
        ptype = 'halos'
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        pn = "particle_position_%s"
        for data_file in sorted(data_files):
            with h5py.File(data_file.filename, "r") as f:
                units = parse_h5_attr(f[pn % "x"], "units")
                pos = data_file._get_particle_positions(ptype, f=f)
                x, y, z = (self.ds.arr(pos[:, i], units)
                           for i in range(3))
                yield "halos", (x, y, z)

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        chunks = list(chunks)
        data_files = set([])
        # Only support halo reading for now.
        assert(len(ptf) == 1)
        assert(list(ptf.keys())[0] == "halos")
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        pn = "particle_position_%s"
        for data_file in sorted(data_files):
            with h5py.File(data_file.filename, "r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    units = parse_h5_attr(f[pn % "x"], "units")
                    pos = data_file._get_particle_positions(ptype, f=f)
                    x, y, z = (self.ds.arr(pos[:, i], units)
                               for i in range(3))
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None: continue
                    for field in field_list:
                        data = f[field][mask].astype("float64")
                        yield (ptype, field), data

    def _initialize_index(self, data_file, regions):
        pcount = data_file.header["num_halos"]
        morton = np.empty(pcount, dtype='uint64')
        mylog.debug("Initializing index % 5i (% 7i particles)",
                    data_file.file_id, pcount)
        ind = 0
        if pcount == 0: return None
        ptype = 'halos'
        with h5py.File(data_file.filename, "r") as f:
            if not f.keys(): return None
            units = parse_h5_attr(f["particle_position_x"], "units")
            pos = data_file._get_particle_positions(ptype, f=f)
            pos = data_file.ds.arr(pos, units). to("code_length")
            dle = self.ds.domain_left_edge.to("code_length")
            dre = self.ds.domain_right_edge.to("code_length")
            if np.any(pos.min(axis=0) < dle) or \
               np.any(pos.max(axis=0) > dre):
                raise YTDomainOverflow(pos.min(axis=0),
                                       pos.max(axis=0),
                                       dle, dre)
            regions.add_data_file(pos, data_file.file_id)
            morton[ind:ind+pos.shape[0]] = compute_morton(
                pos[:,0], pos[:,1], pos[:,2], dle, dre)
        return morton

    def _count_particles(self, data_file):
        return {'halos': data_file.header['num_halos']}

    def _identify_fields(self, data_file):
        with h5py.File(data_file.filename, "r") as f:
            fields = [("halos", field) for field in f]
            units = dict([(("halos", field),
                           parse_h5_attr(f[field], "units"))
                          for field in f])
        return fields, units



"""
Data containers based on geometric selection




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.data_objects.data_containers import \
    YTSelectionContainer0D, YTSelectionContainer1D, \
    YTSelectionContainer2D, YTSelectionContainer3D, YTSelectionContainer
from yt.data_objects.static_output import Dataset
from yt.extern.six import string_types
from yt.funcs import ensure_list, iterable, validate_width_tuple, \
    fix_length, fix_axis, validate_3d_array, validate_float, \
    validate_iterable, validate_object, validate_axis, validate_center
from yt.units.yt_array import \
    YTArray, \
    YTQuantity
from yt.utilities.exceptions import \
    YTSphereTooSmall, \
    YTIllDefinedCutRegion, \
    YTEllipsoidOrdering
from yt.utilities.minimal_representation import \
    MinimalSliceData
from yt.utilities.math_utils import get_rotation_matrix
from yt.utilities.orientation import Orientation
from yt.geometry.selection_routines import points_in_cells
from yt.utilities.on_demand_imports import _scipy


class YTPoint(YTSelectionContainer0D):
    """
    A 0-dimensional object defined by a single point

    Parameters
    ----------
    p: array_like
        A points defined within the domain.  If the domain is
        periodic its position will be corrected to lie inside
        the range [DLE,DRE) to ensure one and only one cell may
        match that point
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
        A dictionary of field parameters than can be accessed by derived
        fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5,0.5,0.5]
    >>> point = ds.point(c)
    """
    _type_name = "point"
    _con_args = ('p',)
    def __init__(self, p, ds=None, field_parameters=None, data_source=None):
        validate_3d_array(p)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        super(YTPoint, self).__init__(ds, field_parameters, data_source)
        if isinstance(p, YTArray):
            # we pass p through ds.arr to ensure code units are attached
            self.p = self.ds.arr(p)
        else:
            self.p = self.ds.arr(p, 'code_length')

class YTOrthoRay(YTSelectionContainer1D):
    """
    This is an orthogonal ray cast through the entire domain, at a specific
    coordinate.

    This object is typically accessed through the `ortho_ray` object that
    hangs off of index objects.  The resulting arrays have their
    dimensionality reduced to one, and an ordered list of points at an
    (x,y) tuple along `axis` are available.

    Parameters
    ----------
    axis : int or char
        The axis along which to slice.  Can be 0, 1, or 2 for x, y, z.
    coords : tuple of floats
        The (plane_x, plane_y) coordinates at which to cast the ray.  Note
        that this is in the plane coordinates: so if you are casting along
        x, this will be (y, z).  If you are casting along y, this will be
        (z, x).  If you are casting along z, this will be (x, y).
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> oray = ds.ortho_ray(0, (0.2, 0.74))
    >>> print oray["Density"]

    Note: The low-level data representation for rays are not guaranteed to be 
    spatially ordered.  In particular, with AMR datasets, higher resolution 
    data is tagged on to the end of the ray.  If you want this data 
    represented in a spatially ordered manner, manually sort it by the "t" 
    field, which is the value of the parametric variable that goes from 0 at 
    the start of the ray to 1 at the end:

    >>> my_ray = ds.ortho_ray(...)
    >>> ray_sort = np.argsort(my_ray["t"])
    >>> density = my_ray["density"][ray_sort]
    """
    _key_fields = ['x','y','z','dx','dy','dz']
    _type_name = "ortho_ray"
    _con_args = ('axis', 'coords')
    def __init__(self, axis, coords, ds=None, 
                 field_parameters=None, data_source=None):
        validate_axis(ds, axis)
        validate_iterable(coords)
        for c in coords:
            validate_float(c)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        super(YTOrthoRay, self).__init__(ds, field_parameters, data_source)
        self.axis = fix_axis(axis, self.ds)
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        self.px_ax = xax
        self.py_ax = yax
        # Even though we may not be using x,y,z we use them here.
        self.px_dx = 'd%s'%('xyz'[self.px_ax])
        self.py_dx = 'd%s'%('xyz'[self.py_ax])
        # Convert coordinates to code length.
        if isinstance(coords[0], YTQuantity):
            self.px = self.ds.quan(coords[0]).to("code_length")
        else:
            self.px = self.ds.quan(coords[0], "code_length")
        if isinstance(coords[1], YTQuantity):
            self.py = self.ds.quan(coords[1]).to("code_length")
        else:
            self.py = self.ds.quan(coords[1], "code_length")
        self.sort_by = 'xyz'[self.axis]

    @property
    def coords(self):
        return (self.px, self.py)

class YTRay(YTSelectionContainer1D):
    """
    This is an arbitrarily-aligned ray cast through the entire domain, at a
    specific coordinate.

    This object is typically accessed through the `ray` object that hangs
    off of index objects.  The resulting arrays have their
    dimensionality reduced to one, and an ordered list of points at an
    (x,y) tuple along `axis` are available, as is the `t` field, which
    corresponds to a unitless measurement along the ray from start to
    end.

    Parameters
    ----------
    start_point : array-like set of 3 floats
        The place where the ray starts.
    end_point : array-like set of 3 floats
        The place where the ray ends.
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> ray = ds.ray((0.2, 0.74, 0.11), (0.4, 0.91, 0.31))
    >>> print ray["Density"], ray["t"], ray["dts"]

    Note: The low-level data representation for rays are not guaranteed to be 
    spatially ordered.  In particular, with AMR datasets, higher resolution 
    data is tagged on to the end of the ray.  If you want this data 
    represented in a spatially ordered manner, manually sort it by the "t" 
    field, which is the value of the parametric variable that goes from 0 at 
    the start of the ray to 1 at the end:

    >>> my_ray = ds.ray(...)
    >>> ray_sort = np.argsort(my_ray["t"])
    >>> density = my_ray["density"][ray_sort]

"""
    _type_name = "ray"
    _con_args = ('start_point', 'end_point')
    _container_fields = ("t", "dts")
    def __init__(self, start_point, end_point, ds=None,
                 field_parameters=None, data_source=None):
        validate_3d_array(start_point)
        validate_3d_array(end_point)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        super(YTRay, self).__init__(ds, field_parameters, data_source)
        if isinstance(start_point, YTArray):
            self.start_point = \
              self.ds.arr(start_point).to("code_length")
        else:
            self.start_point = \
              self.ds.arr(start_point, 'code_length',
                          dtype='float64')
        if isinstance(end_point, YTArray):
            self.end_point = \
              self.ds.arr(end_point).to("code_length")
        else:
            self.end_point = \
              self.ds.arr(end_point, 'code_length',
                          dtype='float64')
        self.vec = self.end_point - self.start_point
        self._set_center(self.start_point)
        self.set_field_parameter('center', self.start_point)
        self._dts, self._ts = None, None

    def _generate_container_field(self, field):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if field == "dts":
            return self._current_chunk.dtcoords
        elif field == "t":
            return self._current_chunk.tcoords
        else:
            raise KeyError(field)

class YTSlice(YTSelectionContainer2D):
    """
    This is a data object corresponding to a slice through the simulation
    domain.

    This object is typically accessed through the `slice` object that hangs
    off of index objects.  Slice is an orthogonal slice through the
    data, taking all the points at the finest resolution available and then
    indexing them.  It is more appropriately thought of as a slice
    'operator' than an object, however, as its field and coordinate can
    both change.

    Parameters
    ----------
    axis : int or char
        The axis along which to slice.  Can be 0, 1, or 2 for x, y, z.
    coord : float
        The coordinate along the axis at which to slice.  This is in
        "domain" coordinates.
    center : array_like, optional
        The 'center' supplied to fields that use it.  Note that this does
        not have to have `coord` as one value.  optional.
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> slice = ds.slice(0, 0.25)
    >>> print slice["Density"]
    """
    _top_node = "/Slices"
    _type_name = "slice"
    _con_args = ('axis', 'coord')
    _container_fields = ("px", "py", "pz", "pdx", "pdy", "pdz")
    def __init__(self, axis, coord, center=None, ds=None,
                 field_parameters=None, data_source=None):
        validate_axis(ds, axis)
        validate_float(coord)
        # center is an optional parameter
        if center is not None:
            validate_center(center)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer2D.__init__(self, axis, ds,
                                        field_parameters, data_source)
        self._set_center(center)
        self.coord = coord

    def _generate_container_field(self, field):
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if field == "px":
            return self._current_chunk.fcoords[:,xax]
        elif field == "py":
            return self._current_chunk.fcoords[:,yax]
        elif field == "pz":
            return self._current_chunk.fcoords[:,self.axis]
        elif field == "pdx":
            return self._current_chunk.fwidth[:,xax] * 0.5
        elif field == "pdy":
            return self._current_chunk.fwidth[:,yax] * 0.5
        elif field == "pdz":
            return self._current_chunk.fwidth[:,self.axis] * 0.5            
        else:
            raise KeyError(field)

    @property
    def _mrep(self):
        return MinimalSliceData(self)

    def hub_upload(self):
        self._mrep.upload()

    def to_pw(self, fields=None, center='c', width=None, origin='center-window'):
        r"""Create a :class:`~yt.visualization.plot_window.PWViewerMPL` from this
        object.

        This is a bare-bones mechanism of creating a plot window from this
        object, which can then be moved around, zoomed, and on and on.  All
        behavior of the plot window is relegated to that routine.
        """
        pw = self._get_pw(fields, center, width, origin, 'Slice')
        return pw

    def plot(self, fields=None):
        if hasattr(self._data_source, "left_edge") and \
            hasattr(self._data_source, "right_edge"):
            left_edge = self._data_source.left_edge
            right_edge = self._data_source.right_edge
            center = (left_edge + right_edge)/2.0
            width = right_edge - left_edge
            xax = self.ds.coordinates.x_axis[self.axis]
            yax = self.ds.coordinates.y_axis[self.axis]
            lx, rx = left_edge[xax], right_edge[xax]
            ly, ry = left_edge[yax], right_edge[yax]
            width = (rx-lx), (ry-ly)
        else:
            width = self.ds.domain_width
            center = self.ds.domain_center
        pw = self._get_pw(fields, center, width, 'native', 'Slice')
        pw.show()
        return pw

class YTCuttingPlane(YTSelectionContainer2D):
    """
    This is a data object corresponding to an oblique slice through the
    simulation domain.

    This object is typically accessed through the `cutting` object
    that hangs off of index objects.  A cutting plane is an oblique
    plane through the data, defined by a normal vector and a coordinate.
    It attempts to guess an 'north' vector, which can be overridden, and
    then it pixelizes the appropriate data onto the plane without
    interpolation.

    Parameters
    ----------
    normal : array_like
        The vector that defines the desired plane.  For instance, the
        angular momentum of a sphere.
    center : array_like
        The center of the cutting plane, where the normal vector is anchored.
    north_vector: array_like, optional
        An optional vector to describe the north-facing direction in the resulting
        plane.
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Notes
    -----

    This data object in particular can be somewhat expensive to create.
    It's also important to note that unlike the other 2D data objects, this
    object provides px, py, pz, as some cells may have a height from the
    plane.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> cp = ds.cutting([0.1, 0.2, -0.9], [0.5, 0.42, 0.6])
    >>> print cp["Density"]
    """
    _plane = None
    _top_node = "/CuttingPlanes"
    _key_fields = YTSelectionContainer2D._key_fields + ['pz','pdz']
    _type_name = "cutting"
    _con_args = ('normal', 'center')
    _tds_attrs = ("_inv_mat",)
    _tds_fields = ("x", "y", "z", "dx")
    _container_fields = ("px", "py", "pz", "pdx", "pdy", "pdz")
    def __init__(self, normal, center, north_vector=None,
                 ds=None, field_parameters=None, data_source=None):
        validate_3d_array(normal)
        validate_center(center)
        if north_vector is not None:
            validate_3d_array(north_vector)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer2D.__init__(self, 4, ds,
                                        field_parameters, data_source)
        self._set_center(center)
        self.set_field_parameter('center',center)
        # Let's set up our plane equation
        # ax + by + cz + d = 0
        self.orienter = Orientation(normal, north_vector = north_vector)
        self._norm_vec = self.orienter.normal_vector
        self._d = -1.0 * np.dot(self._norm_vec, self.center)
        self._x_vec = self.orienter.unit_vectors[0]
        self._y_vec = self.orienter.unit_vectors[1]
        # First we try all three, see which has the best result:
        self._rot_mat = np.array([self._x_vec,self._y_vec,self._norm_vec])
        self._inv_mat = np.linalg.pinv(self._rot_mat)
        self.set_field_parameter('cp_x_vec',self._x_vec)
        self.set_field_parameter('cp_y_vec',self._y_vec)
        self.set_field_parameter('cp_z_vec',self._norm_vec)

    @property
    def normal(self):
        return self._norm_vec

    def _generate_container_field(self, field):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if field == "px":
            x = self._current_chunk.fcoords[:,0] - self.center[0]
            y = self._current_chunk.fcoords[:,1] - self.center[1]
            z = self._current_chunk.fcoords[:,2] - self.center[2]
            tr = np.zeros(x.size, dtype='float64')
            tr = self.ds.arr(tr, "code_length")
            tr += x * self._x_vec[0]
            tr += y * self._x_vec[1]
            tr += z * self._x_vec[2]
            return tr
        elif field == "py":
            x = self._current_chunk.fcoords[:,0] - self.center[0]
            y = self._current_chunk.fcoords[:,1] - self.center[1]
            z = self._current_chunk.fcoords[:,2] - self.center[2]
            tr = np.zeros(x.size, dtype='float64')
            tr = self.ds.arr(tr, "code_length")
            tr += x * self._y_vec[0]
            tr += y * self._y_vec[1]
            tr += z * self._y_vec[2]
            return tr
        elif field == "pz":
            x = self._current_chunk.fcoords[:,0] - self.center[0]
            y = self._current_chunk.fcoords[:,1] - self.center[1]
            z = self._current_chunk.fcoords[:,2] - self.center[2]
            tr = np.zeros(x.size, dtype='float64')
            tr = self.ds.arr(tr, "code_length")
            tr += x * self._norm_vec[0]
            tr += y * self._norm_vec[1]
            tr += z * self._norm_vec[2]
            return tr
        elif field == "pdx":
            return self._current_chunk.fwidth[:,0] * 0.5
        elif field == "pdy":
            return self._current_chunk.fwidth[:,1] * 0.5
        elif field == "pdz":
            return self._current_chunk.fwidth[:,2] * 0.5
        else:
            raise KeyError(field)

    def to_pw(self, fields=None, center='c', width=None, axes_unit=None):
        r"""Create a :class:`~yt.visualization.plot_window.PWViewerMPL` from this
        object.

        This is a bare-bones mechanism of creating a plot window from this
        object, which can then be moved around, zoomed, and on and on.  All
        behavior of the plot window is relegated to that routine.
        """
        normal = self.normal
        center = self.center
        self.fields = ensure_list(fields) + [k for k in self.field_data.keys()
                                             if k not in self._key_fields]
        from yt.visualization.plot_window import get_oblique_window_parameters, PWViewerMPL
        from yt.visualization.fixed_resolution import FixedResolutionBuffer
        (bounds, center_rot) = get_oblique_window_parameters(normal, center, width, self.ds)
        pw = PWViewerMPL(
            self, bounds, fields=self.fields, origin='center-window', 
            periodic=False, oblique=True,
            frb_generator=FixedResolutionBuffer, 
            plot_type='OffAxisSlice')
        if axes_unit is not None:
            pw.set_axes_unit(axes_unit)
        pw._setup_plots()
        return pw

    def to_frb(self, width, resolution, height=None, periodic=False):
        r"""This function returns a FixedResolutionBuffer generated from this
        object.

        An ObliqueFixedResolutionBuffer is an object that accepts a
        variable-resolution 2D object and transforms it into an NxM bitmap that
        can be plotted, examined or processed.  This is a convenience function
        to return an FRB directly from an existing 2D data object.  Unlike the
        corresponding to_frb function for other YTSelectionContainer2D objects, 
        this does not accept a 'center' parameter as it is assumed to be 
        centered at the center of the cutting plane.

        Parameters
        ----------
        width : width specifier
            This can either be a floating point value, in the native domain
            units of the simulation, or a tuple of the (value, unit) style.
            This will be the width of the FRB.
        height : height specifier, optional
            This will be the height of the FRB, by default it is equal to width.
        resolution : int or tuple of ints
            The number of pixels on a side of the final FRB.
        periodic : boolean
            This can be true or false, and governs whether the pixelization
            will span the domain boundaries.

        Returns
        -------
        frb : :class:`~yt.visualization.fixed_resolution.ObliqueFixedResolutionBuffer`
            A fixed resolution buffer, which can be queried for fields.

        Examples
        --------

        >>> v, c = ds.find_max("density")
        >>> sp = ds.sphere(c, (100.0, 'au'))
        >>> L = sp.quantities.angular_momentum_vector()
        >>> cutting = ds.cutting(L, c)
        >>> frb = cutting.to_frb( (1.0, 'pc'), 1024)
        >>> write_image(np.log10(frb["Density"]), 'density_1pc.png')
        """
        if iterable(width):
            validate_width_tuple(width)
            width = self.ds.quan(width[0], width[1])
        if height is None:
            height = width
        elif iterable(height):
            validate_width_tuple(height)
            height = self.ds.quan(height[0], height[1])
        if not iterable(resolution):
            resolution = (resolution, resolution)
        from yt.visualization.fixed_resolution import FixedResolutionBuffer
        bounds = (-width/2.0, width/2.0, -height/2.0, height/2.0)
        frb = FixedResolutionBuffer(self, bounds, resolution,
                                           periodic=periodic)
        return frb

class YTDisk(YTSelectionContainer3D):
    """
    By providing a *center*, a *normal*, a *radius* and a *height* we
    can define a cylinder of any proportion.  Only cells whose centers are
    within the cylinder will be selected.

    Parameters
    ----------
    center : array_like
        coordinate to which the normal, radius, and height all reference
    normal : array_like
        the normal vector defining the direction of lengthwise part of the 
        cylinder
    radius : float
        the radius of the cylinder
    height : float
        the distance from the midplane of the cylinder to the top and 
        bottom planes
    fields : array of fields, optional
        any fields to be pre-loaded in the cylinder object
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5,0.5,0.5]
    >>> disk = ds.disk(c, [1,0,0], (1, 'kpc'), (10, 'kpc'))
    """
    _type_name = "disk"
    _con_args = ('center', '_norm_vec', 'radius', 'height')
    def __init__(self, center, normal, radius, height, fields=None,
                 ds=None, field_parameters=None, data_source=None):
        validate_center(center)
        validate_3d_array(normal)
        validate_float(radius)
        validate_float(height)
        validate_iterable(fields)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, center, ds,
                                        field_parameters, data_source)
        self._norm_vec = np.array(normal)/np.sqrt(np.dot(normal,normal))
        self.set_field_parameter("normal", self._norm_vec)
        self.set_field_parameter("center", self.center)
        self.height = fix_length(height, self.ds)
        self.radius = fix_length(radius, self.ds)
        self._d = -1.0 * np.dot(self._norm_vec, self.center)

class YTRegion(YTSelectionContainer3D):
    """A 3D region of data with an arbitrary center.

    Takes an array of three *left_edge* coordinates, three
    *right_edge* coordinates, and a *center* that can be anywhere
    in the domain. If the selected region extends past the edges
    of the domain, no data will be found there, though the
    object's `left_edge` or `right_edge` are not modified.

    Parameters
    ----------
    center : array_like
        The center of the region
    left_edge : array_like
        The left edge of the region
    right_edge : array_like
        The right edge of the region
    """
    _type_name = "region"
    _con_args = ('center', 'left_edge', 'right_edge')
    def __init__(self, center, left_edge, right_edge, fields=None,
                 ds=None, field_parameters=None, data_source=None):
        if center is not None:
            validate_center(center)
        validate_3d_array(left_edge)
        validate_3d_array(right_edge)
        validate_iterable(fields)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, center, ds,
                                        field_parameters, data_source)
        if not isinstance(left_edge, YTArray):
            self.left_edge = self.ds.arr(left_edge, 'code_length')
        else:
            # need to assign this dataset's unit registry to the YTArray
            self.left_edge = self.ds.arr(left_edge.copy())
        if not isinstance(right_edge, YTArray):
            self.right_edge = self.ds.arr(right_edge, 'code_length')
        else:
            # need to assign this dataset's unit registry to the YTArray
            self.right_edge = self.ds.arr(right_edge.copy())

class YTDataCollection(YTSelectionContainer3D):
    """
    By selecting an arbitrary *object_list*, we can act on those grids.
    Child cells are not returned.
    """
    _type_name = "data_collection"
    _con_args = ("_obj_list",)
    def __init__(self, obj_list, ds=None, field_parameters=None,
                 data_source=None, center=None):
        validate_iterable(obj_list)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        if center is not None:
            validate_center(center)
        YTSelectionContainer3D.__init__(self, center, ds,
                                        field_parameters, data_source)
        self._obj_ids = np.array([o.id - o._id_offset for o in obj_list],
                                dtype="int64")
        self._obj_list = obj_list

class YTSphere(YTSelectionContainer3D):
    """
    A sphere of points defined by a *center* and a *radius*.

    Parameters
    ----------
    center : array_like
        The center of the sphere.
    radius : float, width specifier, or YTQuantity
        The radius of the sphere. If passed a float,
        that will be interpreted in code units. Also
        accepts a (radius, unit) tuple or YTQuantity
        instance with units attached.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5,0.5,0.5]
    >>> sphere = ds.sphere(c, (1., "kpc"))
    """
    _type_name = "sphere"
    _con_args = ('center', 'radius')
    def __init__(self, center, radius, ds=None,
                 field_parameters=None, data_source=None):
        validate_center(center)
        validate_float(radius)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        super(YTSphere, self).__init__(center, ds,
                                           field_parameters, data_source)
        # Unpack the radius, if necessary
        radius = fix_length(radius, self.ds)
        if radius < self.index.get_smallest_dx():
            raise YTSphereTooSmall(ds, radius.in_units("code_length"),
                                   self.index.get_smallest_dx().in_units("code_length"))
        self.set_field_parameter('radius', radius)
        self.set_field_parameter("center", self.center)
        self.radius = radius

class YTEllipsoid(YTSelectionContainer3D):
    """
    By providing a *center*,*A*,*B*,*C*,*e0*,*tilt* we
    can define a ellipsoid of any proportion.  Only cells whose
    centers are within the ellipsoid will be selected.

    Parameters
    ----------
    center : array_like
        The center of the ellipsoid.
    A : float
        The magnitude of the largest axis (semi-major) of the ellipsoid.
    B : float
        The magnitude of the medium axis (semi-medium) of the ellipsoid.
    C : float
        The magnitude of the smallest axis (semi-minor) of the ellipsoid.
    e0 : array_like (automatically normalized)
        the direction of the largest semi-major axis of the ellipsoid
    tilt : float
        After the rotation about the z-axis to allign e0 to x in the x-y
        plane, and then rotating about the y-axis to align e0 completely
        to the x-axis, tilt is the angle in radians remaining to
        rotate about the x-axis to align both e1 to the y-axis and e2 to
        the z-axis.
    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5,0.5,0.5]
    >>> ell = ds.ellipsoid(c, 0.1, 0.1, 0.1, np.array([0.1, 0.1, 0.1]), 0.2)
    """
    _type_name = "ellipsoid"
    _con_args = ('center', '_A', '_B', '_C', '_e0', '_tilt')
    def __init__(self, center, A, B, C, e0, tilt, fields=None,
                 ds=None, field_parameters=None, data_source=None):
        validate_center(center)
        validate_float(A)
        validate_float(B)
        validate_float(C)
        validate_3d_array(e0)
        validate_float(tilt)
        validate_iterable(fields)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, center, ds,
                                        field_parameters, data_source)
        # make sure the magnitudes of semi-major axes are in order
        if A<B or B<C:
            raise YTEllipsoidOrdering(ds, A, B, C)
        # make sure the smallest side is not smaller than dx
        self._A = self.ds.quan(A, 'code_length')
        self._B = self.ds.quan(B, 'code_length')
        self._C = self.ds.quan(C, 'code_length')
        if self._C < self.index.get_smallest_dx():
            raise YTSphereTooSmall(self.ds, self._C, self.index.get_smallest_dx())
        self._e0 = e0 = e0 / (e0**2.0).sum()**0.5
        self._tilt = tilt
 
        # find the t1 angle needed to rotate about z axis to align e0 to x
        t1 = np.arctan(e0[1] / e0[0])
        # rotate e0 by -t1
        RZ = get_rotation_matrix(t1, (0,0,1)).transpose()
        r1 = (e0 * RZ).sum(axis = 1)
        # find the t2 angle needed to rotate about y axis to align e0 to x
        t2 = np.arctan(-r1[2] / r1[0])
        """
        calculate the original e1
        given the tilt about the x axis when e0 was aligned 
        to x after t1, t2 rotations about z, y
        """
        RX = get_rotation_matrix(-tilt, (1, 0, 0)).transpose()
        RY = get_rotation_matrix(-t2,   (0, 1, 0)).transpose()
        RZ = get_rotation_matrix(-t1,   (0, 0, 1)).transpose()
        e1 = ((0, 1, 0) * RX).sum(axis=1)
        e1 = (e1 * RY).sum(axis=1)
        e1 = (e1 * RZ).sum(axis=1)
        e2 = np.cross(e0, e1)

        self._e1 = e1
        self._e2 = e2

        self.set_field_parameter('A', A)
        self.set_field_parameter('B', B)
        self.set_field_parameter('C', C)
        self.set_field_parameter('e0', e0)
        self.set_field_parameter('e1', e1)
        self.set_field_parameter('e2', e2)

class YTCutRegion(YTSelectionContainer3D):
    """
    This is a data object designed to allow individuals to apply logical
    operations to fields and filter as a result of those cuts.

    Parameters
    ----------
    data_source : YTSelectionContainer3D
        The object to which cuts will be applied.
    conditionals : list of strings
        A list of conditionals that will be evaluated.  In the namespace
        available, these conditionals will have access to 'obj' which is a data
        object of unknown shape, and they must generate a boolean array.  For
        instance, conditionals = ["obj['temperature'] < 1e3"]

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> sp = ds.sphere("max", (1.0, 'Mpc'))
    >>> cr = ds.cut_region(sp, ["obj['temperature'] < 1e3"])
    """
    _type_name = "cut_region"
    _con_args = ("base_object", "conditionals")
    def __init__(self, data_source, conditionals, ds=None,
                 field_parameters=None, base_object=None,
                 locals={}):
        validate_object(data_source, YTSelectionContainer)
        validate_iterable(conditionals)
        for condition in conditionals:
            validate_object(condition, string_types)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(base_object, YTSelectionContainer)
        if base_object is not None:
            # passing base_object explicitly has been deprecated,
            # but we handle it here for backward compatibility
            if data_source is not None:
                raise RuntimeError(
                    "Cannot use both base_object and data_source")
            data_source=base_object

        self.conditionals = ensure_list(conditionals)
        if isinstance(data_source, YTCutRegion):
            # If the source is also a cut region, add its conditionals
            # and set the source to be its source.
            # Preserve order of conditionals.
            self.conditionals = data_source.conditionals + \
              self.conditionals
            data_source = data_source.base_object

        super(YTCutRegion, self).__init__(
            data_source.center, ds, field_parameters, data_source=data_source)
        self.base_object = data_source
        self.locals = locals
        self._selector = None
        # Need to interpose for __getitem__, fwidth, fcoords, icoords, iwidth,
        # ires and get_data

    def chunks(self, fields, chunking_style, **kwargs):
        # We actually want to chunk the sub-chunk, not ourselves.  We have no
        # chunks to speak of, as we do not data IO.
        for chunk in self.index._chunk(self.base_object,
                                       chunking_style,
                                       **kwargs):
            with self.base_object._chunked_read(chunk):
                with self._chunked_read(chunk):
                    self.get_data(fields)
                    yield self

    def get_data(self, fields = None):
        fields = ensure_list(fields)
        self.base_object.get_data(fields)
        ind = self._cond_ind
        for field in fields:
            f = self.base_object[field]
            if f.shape != ind.shape:
                parent = getattr(self, "parent", self.base_object)
                self.field_data[field] = \
                  parent[field][self._part_ind(field[0])]
            else:
                self.field_data[field] = self.base_object[field][ind]

    @property
    def blocks(self):
        # We have to take a slightly different approach here.  Note that all
        # that .blocks has to yield is a 3D array and a mask.
        for obj, m in self.base_object.blocks:
            m = m.copy()
            with obj._field_parameter_state(self.field_parameters):
                for cond in self.conditionals:
                    ss = eval(cond)
                    m = np.logical_and(m, ss, m)
            if not np.any(m): continue
            yield obj, m

    @property
    def _cond_ind(self):
        ind = None
        obj = self.base_object
        locals = self.locals.copy()
        if 'obj' in locals:
            raise RuntimeError('"obj" has been defined in the "locals" ; this is not supported, please rename the variable.')
        locals['obj'] = obj
        with obj._field_parameter_state(self.field_parameters):
            for cond in self.conditionals:
                res = eval(cond, locals)
                if ind is None: ind = res
                if ind.shape != res.shape:
                    raise YTIllDefinedCutRegion(self.conditionals)
                np.logical_and(res, ind, ind)
        return ind

    def _part_ind_KDTree(self, ptype):
        '''Find the particles in cells using a KDTree approach.'''
        parent = getattr(self, "parent", self.base_object)
        units = "code_length"

        pos = np.stack([self[("index", 'x')].to(units),
                        self[("index", 'y')].to(units),
                        self[("index", 'z')].to(units)], axis=1).value
        dx = np.stack([self[("index", "dx")].to(units),
                       self[("index", "dy")].to(units),
                       self[("index", "dz")].to(units)], axis=1).value
        ppos = np.stack([parent[(ptype, "particle_position_x")],
                         parent[(ptype, "particle_position_y")],
                         parent[(ptype, "particle_position_z")]], axis=1).value
        levels = self[("index", "grid_level")].astype('int32').value
        levelmin = levels.min()
        levelmax = levels.max()

        mask = np.zeros(ppos.shape[0], dtype=bool)

        for lvl in range(levelmax, levelmin-1, -1):
            # Filter out cells not in the current level
            lvl_mask = (levels == lvl)
            dx_loc = dx[lvl_mask]
            pos_loc = pos[lvl_mask]

            grid_tree = _scipy.spatial.cKDTree(pos_loc, boxsize=1)

            # Compute closest cell for all remaining particles
            dist, icell = grid_tree.query(ppos[~mask], distance_upper_bound=dx_loc.max(),
                                          p=np.inf)
            mask_loc = np.isfinite(dist[:])

            # Check that particles within dx of a cell are in it
            i = icell[mask_loc]
            dist = np.abs(ppos[~mask][mask_loc, :] - pos_loc[i])
            tmp_mask = np.all(dist <= (dx_loc[i] / 2), axis=1)

            mask_loc[mask_loc] = tmp_mask

            # Update the particle mask with particles found at this level
            mask[~mask] |= mask_loc

        return mask

    def _part_ind_brute_force(self, ptype):
        parent = getattr(self, "parent", self.base_object)
        units = "code_length"
        mask = points_in_cells(
            self[("index", "x")].to(units),
            self[("index", "y")].to(units),
            self[("index", "z")].to(units),
            self[("index", "dx")].to(units),
            self[("index", "dy")].to(units),
            self[("index", "dz")].to(units),
            parent[(ptype, "particle_position_x")].to(units),
            parent[(ptype, "particle_position_y")].to(units),
            parent[(ptype, "particle_position_z")].to(units))

        return mask

    def _part_ind(self, ptype):
        # If scipy is installed, use the fast KD tree
        # implementation. Else, fall back onto the direct
        # brute-force algorithm.
        try:
            _scipy.spatial.KDTree
            return self._part_ind_KDTree(ptype)
        except ImportError:
            return self._part_ind_brute_force(ptype)

    @property
    def icoords(self):
        return self.base_object.icoords[self._cond_ind,:]

    @property
    def fcoords(self):
        return self.base_object.fcoords[self._cond_ind,:]

    @property
    def ires(self):
        return self.base_object.ires[self._cond_ind]

    @property
    def fwidth(self):
        return self.base_object.fwidth[self._cond_ind,:]

class YTIntersectionContainer3D(YTSelectionContainer3D):
    """
    This is a more efficient method of selecting the intersection of multiple
    data selection objects.

    Creating one of these objects returns the intersection of all of the
    sub-objects; it is designed to be a faster method than chaining & ("and")
    operations to create a single, large intersection.

    Parameters
    ----------
    data_objects : Iterable of YTSelectionContainer
        The data objects to intersect

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> sp1 = ds.sphere((0.4, 0.5, 0.6), 0.15)
    >>> sp2 = ds.sphere((0.38, 0.51, 0.55), 0.1)
    >>> sp3 = ds.sphere((0.35, 0.5, 0.6), 0.15)
    >>> new_obj = ds.intersection((sp1, sp2, sp3))
    >>> print(new_obj.sum("cell_volume"))
    """
    _type_name = "intersection"
    _con_args = ("data_objects",)
    def __init__(self, data_objects, ds = None, field_parameters = None,
                 data_source = None):
        validate_iterable(data_objects)
        for obj in data_objects:
            validate_object(obj, YTSelectionContainer)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, None, ds, field_parameters,
                data_source)
        # ensure_list doesn't check for tuples
        if isinstance(data_objects, tuple):
            data_objects = list(data_objects)
        self.data_objects = ensure_list(data_objects)

class YTDataObjectUnion(YTSelectionContainer3D):
    """
    This is a more efficient method of selecting the union of multiple
    data selection objects.

    Creating one of these objects returns the union of all of the sub-objects;
    it is designed to be a faster method than chaining | (or) operations to
    create a single, large union.

    Parameters
    ----------
    data_objects : Iterable of YTSelectionContainer
        The data objects to union

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> sp1 = ds.sphere((0.4, 0.5, 0.6), 0.1)
    >>> sp2 = ds.sphere((0.3, 0.5, 0.15), 0.1)
    >>> sp3 = ds.sphere((0.5, 0.5, 0.9), 0.1)
    >>> new_obj = ds.union((sp1, sp2, sp3))
    >>> print(new_obj.sum("cell_volume"))
    """
    _type_name = "union"
    _con_args = ("data_objects",)
    def __init__(self, data_objects, ds = None, field_parameters = None,
                 data_source = None):
        validate_iterable(data_objects)
        for obj in data_objects:
            validate_object(obj, YTSelectionContainer)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, None, ds, field_parameters,
                data_source)
        # ensure_list doesn't check for tuples
        if isinstance(data_objects, tuple):
            data_objects = list(data_objects)
        self.data_objects = ensure_list(data_objects)



from yt.testing import \
    assert_array_equal, \
    fake_amr_ds, \
    fake_random_ds
from yt.units import cm

def test_box_creation():

    # test that creating a region with left and right edge
    # with units works
    ds = fake_random_ds(32, length_unit=2)
    reg = ds.box([0, 0, 0]*cm, [2, 2, 2]*cm)
    dens_units = reg['density']

    reg = ds.box([0, 0, 0], [1, 1, 1])
    dens_no_units = reg['density']

    assert_array_equal(dens_units, dens_no_units)

def test_max_level_min_level_semantics():
    ds = fake_amr_ds()
    ad = ds.all_data()
    assert ad['grid_level'].max() == 4
    ad.max_level = 2
    assert ad['grid_level'].max() == 2
    ad.max_level = 8
    assert ad['grid_level'].max() == 4
    ad.min_level = 2
    assert ad['grid_level'].min() == 2
    ad.min_level = 0
    assert ad['grid_level'].min() == 0
